"""xorq expression API definitions."""

from __future__ import annotations

import functools
from pathlib import Path
from typing import TYPE_CHECKING, Any, Mapping

import pyarrow as pa
import toolz
from opentelemetry import trace

import xorq.vendor.ibis.expr.types as ir
from xorq.backends.xorq import Backend
from xorq.common.exceptions import XorqError
from xorq.common.utils.caching_utils import find_backend
from xorq.common.utils.defer_utils import (  # noqa: F403
    deferred_read_csv,
    deferred_read_parquet,
    rbr_wrapper,
)
from xorq.common.utils.io_utils import (
    extract_suffix,
    maybe_open,
)
from xorq.common.utils.otel_utils import tracer
from xorq.common.utils.rbr_utils import otel_instrument_reader
from xorq.expr.ml import (
    calc_split_column,
    train_test_splits,
)
from xorq.expr.relations import (
    CachedNode,
    Tag,
    register_and_transform_remote_tables,
)
from xorq.vendor.ibis.expr import api
from xorq.vendor.ibis.expr.api import *  # noqa: F403
from xorq.vendor.ibis.expr.sql import SQLString


if TYPE_CHECKING:
    from collections.abc import Sequence
    from io import TextIOWrapper
    from pathlib import Path

    import pandas as pd
    import pyarrow as pa


__all__ = (
    "execute",
    "calc_split_column",
    "read_csv",
    "read_parquet",
    "read_postgres",
    "read_pyarrow_stream",
    "register",
    "train_test_splits",
    "to_parquet",
    "to_csv",
    "to_json",
    "to_pyarrow",
    "to_pyarrow_batches",
    "to_pyarrow_stream",
    "to_sql",
    "get_plans",
    "deferred_read_csv",
    "deferred_read_parquet",
    "get_object_metadata",
    *api.__all__,
)


def read_pyarrow_stream(
    source,
    con=None,
    table_name=None,
    **kwargs,
) -> ir.Table:
    from xorq.config import _backend_init

    con = con or _backend_init()
    rbr = pa.ipc.open_stream(source, **kwargs)
    return con.read_record_batches(rbr, table_name=table_name)


def read_csv(
    sources: str | Path | Sequence[str | Path],
    table_name: str | None = None,
    **kwargs: Any,
) -> ir.Table:
    """Lazily load a CSV or set of CSVs.

    This function delegates to the `read_csv` method on the current default
    backend (DuckDB or `xorq.config.default_backend`).

    Parameters
    ----------
    sources
        A filesystem path or URL or list of same.  Supports CSV and TSV files.
    table_name
        A name to refer to the table.  If not provided, a name will be generated.
    kwargs
        Backend-specific keyword arguments for the file type. For the DuckDB
        backend used by default, please refer to:

        * CSV/TSV: https://duckdb.org/docs/data/csv/overview.html#parameters.

    Returns
    -------
    ir.Table
        Table expression representing a file

    Examples
    --------
    >>> import xorq.api as xo
    >>> xo.options.interactive = True
    >>> lines = '''a,b
    ... 1,d
    ... 2,
    ... ,f
    ... '''
    >>> with open("/tmp/lines.csv", mode="w") as f:
    ...     nbytes = f.write(lines)  # nbytes is unused
    >>> t = xo.read_csv("/tmp/lines.csv")
    >>> t
    ┏━━━━━━━┳━━━━━━━━┓
    ┃ a     ┃ b      ┃
    ┡━━━━━━━╇━━━━━━━━┩
    │ int64 │ string │
    ├───────┼────────┤
    │     1 │ d      │
    │     2 │ NULL   │
    │  NULL │ f      │
    └───────┴────────┘

    """
    from xorq.config import _backend_init

    con = _backend_init()
    return con.read_csv(sources, table_name=table_name, **kwargs)


def read_parquet(
    sources: str | Path | Sequence[str | Path],
    table_name: str | None = None,
    **kwargs: Any,
) -> ir.Table:
    """Lazily load a parquet file or set of parquet files.

    This function delegates to the `read_parquet` method on the current default
    backend (DuckDB or `ibis.config.default_backend`).

    Parameters
    ----------
    sources
        A filesystem path or URL or list of same.
    table_name
        A name to refer to the table.  If not provided, a name will be generated.
    kwargs
        Backend-specific keyword arguments for the file type. For the DuckDB
        backend used by default, please refer to:

        * Parquet: https://duckdb.org/docs/data/parquet

    Returns
    -------
    ir.Table
        Table expression representing a file

    Examples
    --------
    >>> import xorq.api as xo
    >>> import pandas as pd
    >>> xo.options.interactive = True
    >>> df = pd.DataFrame({"a": [1, 2, 3], "b": list("ghi")})
    >>> df
       a  b
    0  1  g
    1  2  h
    2  3  i
    >>> df.to_parquet("/tmp/data.parquet")
    >>> t = xo.read_parquet("/tmp/data.parquet")
    >>> t
    ┏━━━━━━━┳━━━━━━━━┓
    ┃ a     ┃ b      ┃
    ┡━━━━━━━╇━━━━━━━━┩
    │ int64 │ string │
    ├───────┼────────┤
    │     1 │ g      │
    │     2 │ h      │
    │     3 │ i      │
    └───────┴────────┘

    """
    from xorq.config import _backend_init

    con = _backend_init()
    return con.read_parquet(sources, table_name=table_name, **kwargs)


def register(
    source: str | Path | pa.Table | pa.RecordBatch | pa.Dataset | pd.DataFrame,
    table_name: str | None = None,
    **kwargs: Any,
):
    from xorq.config import _backend_init

    con = _backend_init()
    return con.register(source, table_name=table_name, **kwargs)


def read_postgres(
    uri: str,
    table_name: str | None = None,
    **kwargs: Any,
):
    from xorq.config import _backend_init

    con = _backend_init()
    return con.read_postgres(uri, table_name=table_name, **kwargs)


@functools.cache
def _cached_with_op(op, pretty, compiler):
    expr = op.to_expr()
    sg_expr = compiler.to_sqlglot(expr)
    sql = sg_expr.sql(dialect=compiler.dialect, pretty=pretty)
    return sql


get_compiler = toolz.excepts(
    (XorqError, AttributeError),
    lambda e: e._find_backend(use_default=True).compiler,
    lambda _: Backend.compiler,
)


def to_sql(expr: ir.Expr, compiler=None, pretty: bool = True) -> SQLString:
    """Return the formatted SQL string for an expression.

    Parameters
    ----------
    expr
        Ibis expression.
    compiler
        The target compiler to use to translate the Ibis expr
    pretty
        Whether to use pretty formatting.

    Returns
    -------
    str
        Formatted SQL string

    """

    assert isinstance(expr, ir.Expr)

    if compiler is None:
        compiler = get_compiler(expr)

    unbound = _remove_tag_nodes(expr).unbind().op()
    return SQLString(_cached_with_op(unbound, pretty, compiler))


@tracer.start_as_current_span("_register_and_transform_cache_tables")
def _register_and_transform_cache_tables(expr):
    """This function will sequentially execute any cache node that is not already cached"""

    def fn(node, kwargs):
        if kwargs:
            node = node.__recreate__(kwargs)
        if isinstance(node, CachedNode):
            uncached, cache = node.parent, node.cache
            node = cache.set_default(uncached, uncached.op())
        return node

    op = expr.op()
    out = op.replace(fn)

    return out.to_expr()


@tracer.start_as_current_span("_transform_deferred_reads")
def _transform_deferred_reads(expr):
    dt_to_read = {}

    span = trace.get_current_span()

    def replace_read(node, _kwargs):
        from xorq.expr.relations import Read

        if isinstance(node, Read):
            read_kwargs = dict(node.read_kwargs)
            span.add_event(
                "replace_read",
                {
                    "engine": node.source.name,
                    "method_name": node.method_name,
                    "path": str(
                        read_kwargs.get("path")
                        or read_kwargs.get("source")
                        or read_kwargs.get("source_list")
                    ),
                },
            )
            # FIXME: pandas read is not lazy, leave it to the pandas executor to do
            node = dt_to_read[node] = node.make_dt()
        else:
            if _kwargs:
                node = node.__recreate__(_kwargs)
        return node

    expr = expr.op().replace(replace_read).to_expr()
    return expr, dt_to_read


@tracer.start_as_current_span("execute")
def execute(expr: ir.Expr, **kwargs: Any):
    """Execute an expression against its backend if one exists.

    Parameters
    ----------
    kwargs
        Keyword arguments

    Examples
    --------
    >>> import xorq.api as xo
    >>> t = xo.examples.penguins.fetch()
    >>> t.execute()
           species     island  bill_length_mm  ...  body_mass_g     sex  year
    0       Adelie  Torgersen            39.1  ...       3750.0    male  2007
    1       Adelie  Torgersen            39.5  ...       3800.0  female  2007
    2       Adelie  Torgersen            40.3  ...       3250.0  female  2007
    3       Adelie  Torgersen             NaN  ...          NaN    None  2007
    4       Adelie  Torgersen            36.7  ...       3450.0  female  2007
    ..         ...        ...             ...  ...          ...     ...   ...
    339  Chinstrap      Dream            55.8  ...       4000.0    male  2009
    340  Chinstrap      Dream            43.5  ...       3400.0  female  2009
    341  Chinstrap      Dream            49.6  ...       3775.0    male  2009
    342  Chinstrap      Dream            50.8  ...       4100.0    male  2009
    343  Chinstrap      Dream            50.2  ...       3775.0  female  2009
    [344 rows x 8 columns]

    Scalar parameters can be supplied dynamically during execution.
    >>> species = xo.param("string")
    >>> expr = t.filter(t.species == species).order_by(t.bill_length_mm)
    >>> expr.execute(limit=3, params={species: "Gentoo"})
      species  island  bill_length_mm  ...  body_mass_g     sex  year
    0  Gentoo  Biscoe            40.9  ...         4650  female  2007
    1  Gentoo  Biscoe            41.7  ...         4700  female  2009
    2  Gentoo  Biscoe            42.0  ...         4150  female  2007
    <BLANKLINE>
    [3 rows x 8 columns]
    """

    if (con := expr._find_backend(use_default=True)).name == "pandas":
        return _pandas_execute(con, expr, **kwargs)

    batch_reader = to_pyarrow_batches(expr, **kwargs)
    with tracer.start_as_current_span("read_pandas"):
        df = batch_reader.read_pandas(timestamp_as_object=True)
    return expr.__pandas_result__(df)


@tracer.start_as_current_span("_remove_tag_nodes")
def _remove_tag_nodes(expr):
    from xorq.common.utils.graph_utils import replace_nodes

    def replacer(node, kwargs):
        if isinstance(node, Tag):
            while isinstance(node, Tag):
                node = node.parent
            node = replace_nodes(replacer, node)
        elif kwargs:
            node = node.__recreate__(kwargs)
        return node

    return replace_nodes(replacer, expr).to_expr()


@tracer.start_as_current_span("_transform_expr")
def _transform_expr(expr, **kwargs):
    expr = _remove_tag_nodes(expr)
    expr = _register_and_transform_cache_tables(expr)
    expr, created = register_and_transform_remote_tables(expr, **kwargs)
    expr, dt_to_read = _transform_deferred_reads(expr)
    return (expr, created)


def _pandas_execute(con, expr: ir.Expr, **kwargs):
    from xorq.expr.relations import FlightExpr, FlightUDXF

    span = trace.get_current_span()

    node = expr.op()
    if isinstance(node, (FlightExpr, FlightUDXF)):
        # TODO: verify correct caching behavior
        span.set_attribute("engine", "flight")
        df = node.to_rbr().read_pandas(timestamp_as_object=True)
        return expr.__pandas_result__(df)
    (expr, created) = _transform_expr(expr)

    span.set_attribute("engine", "pandas")
    return con.execute(expr, **kwargs)


@tracer.start_as_current_span("to_pyarrow_batches")
def to_pyarrow_batches(
    expr: ir.Expr,
    *,
    chunk_size: int = 1_000_000,
    **kwargs: Any,
):
    """Execute expression and return a RecordBatchReader.

    This method is eager and will execute the associated expression
    immediately.

    Parameters
    ----------
    chunk_size
        Maximum number of rows in each returned record batch.
    kwargs
        Keyword arguments

    Returns
    -------
    results
        RecordBatchReader
    """
    from xorq.expr.relations import FlightExpr, FlightUDXF

    span = trace.get_current_span()

    if isinstance(expr.op(), (FlightExpr, FlightUDXF)):
        # TODO: verify correct caching behavior
        span.set_attribute("engine", "flight")
        return expr.op().to_rbr()
    (expr, created) = _transform_expr(expr)
    con, _ = find_backend(expr.op(), use_default=True)

    span.set_attribute("engine", con.name)
    reader = con.to_pyarrow_batches(expr, chunk_size=chunk_size, **kwargs)

    def clean_up():
        for table_name, conn in created.items():
            try:
                conn.drop_table(table_name, force=True)
            except Exception:
                conn.drop_view(table_name)

    return otel_instrument_reader(rbr_wrapper(reader, clean_up))


def to_pyarrow(expr: ir.Expr, **kwargs: Any):
    """Execute expression and return results in as a pyarrow table.

    This method is eager and will execute the associated expression
    immediately.

    Parameters
    ----------
    kwargs
        Keyword arguments

    Returns
    -------
    Table
        A pyarrow table holding the results of the executed expression.
    """
    batch_reader = to_pyarrow_batches(expr, **kwargs)
    arrow_table = batch_reader.read_all()
    return expr.__pyarrow_result__(arrow_table)


def to_pyarrow_stream(expr: ir.Expr, sink: Any, **kwargs: Any):
    rbr = expr.to_pyarrow_batches()
    with maybe_open(sink, "wb") as fh:
        try:
            writer = pa.ipc.new_stream(fh, rbr.schema, **kwargs)
            for batch in rbr:
                writer.write_batch(batch)
        finally:
            writer.close()


def to_parquet(
    expr: ir.Expr,
    path: str | Path,
    params: Mapping[ir.Scalar, Any] | None = None,
    **kwargs: Any,
):
    """Write the results of executing the given expression to a parquet file.

    This method is eager and will execute the associated expression
    immediately.

    See https://arrow.apache.org/docs/python/generated/pyarrow.parquet.ParquetWriter.html for details.

    Parameters
    ----------
    path
        A string or Path where the Parquet file will be written.
    params
        Mapping of scalar parameter expressions to value.
    **kwargs
        Additional keyword arguments passed to pyarrow.parquet.ParquetWriter

    Examples
    --------
    Write out an expression to a single parquet file.

    >>> import ibis
    >>> import tempfile
    >>> penguins = ibis.examples.penguins.fetch()
    >>> penguins.to_parquet(tempfile.mktemp())
    """
    import pyarrow  # noqa: ICN001, F401
    import pyarrow.parquet as pq
    import pyarrow_hotfix  # noqa: F401

    with to_pyarrow_batches(expr, params=params) as batch_reader:
        with pq.ParquetWriter(path, batch_reader.schema, **kwargs) as writer:
            for batch in batch_reader:
                writer.write_batch(batch)


def to_csv(
    expr: ir.Expr,
    path: str | Path,
    params: Mapping[ir.Scalar, Any] | None = None,
    **kwargs: Any,
):
    """Write the results of executing the given expression to a CSV file.

    This method is eager and will execute the associated expression
    immediately.

    Parameters
    ----------
    path
        The data source. A string or Path to the CSV file.
    params
        Mapping of scalar parameter expressions to value.
    **kwargs
        Additional keyword arguments passed to pyarrow.csv.CSVWriter

    https://arrow.apache.org/docs/python/generated/pyarrow.csv.CSVWriter.htmlditional keyword arguments passed to pyarrow.csv.CSVWriter
    """

    import pyarrow  # noqa: ICN001, F401
    import pyarrow.csv as pcsv
    import pyarrow_hotfix  # noqa: F401

    with pcsv.CSVWriter(path, schema=expr.schema().to_pyarrow(), **kwargs) as writer:
        with to_pyarrow_batches(expr, params=params) as batch_reader:
            for batch in batch_reader:
                writer.write_batch(batch)


def to_json(
    expr: ir.Expr,
    path: str | Path | TextIOWrapper,
    params: Mapping[ir.Scalar, Any] | None = None,
):
    """Write the results of `expr` to a NDJSON file.

    This method is eager and will execute the associated expression
    immediately.

    Parameters
    ----------
    path
        The data source. A string or Path to the Delta Lake table.
    **kwargs
        Additional, backend-specific keyword arguments.

    https://github.com/ndjson/ndjson-spec
    """
    import pyarrow  # noqa: ICN001, F401
    import pyarrow_hotfix  # noqa: F401

    from xorq.common.utils.io_utils import maybe_open

    with maybe_open(path, "w") as f:
        with to_pyarrow_batches(expr, params=params) as batch_reader:
            for batch in batch_reader:
                df = batch.to_pandas()
                batch_json = df.to_json(orient="records", lines=True)
                f.write(batch_json)


def get_plans(expr):
    _expr, _ = _transform_expr(expr)
    con, _ = find_backend(_expr.op())
    sql = f"EXPLAIN {to_sql(_expr)}"
    return con.con.sql(sql).to_pandas().set_index("plan_type")["plan"].to_dict()


def get_object_metadata(path: str, **kwargs: Any) -> dict:
    from xorq.config import _backend_init

    con = _backend_init()

    suffix = extract_suffix(path).lstrip(".")

    if "storage_options" in kwargs:
        kwargs["storage_options"] = dict(kwargs.pop("storage_options"))

    return con.con.get_object_metadata(path, suffix, **kwargs)



import datetime
import functools
import json
import operator
import time
from pathlib import Path
from queue import (
    Empty as QueueEmpty,
)
from queue import (
    Queue,
)
from threading import (
    Thread,
)

import toolz
from attr import (
    field,
    frozen,
)
from attr.validators import (
    deep_iterable,
    instance_of,
    is_callable,
    optional,
)
from toolz import (
    compose,
    identity,
)
from toolz.curried import (
    map as cmap,
)

from xorq.common.utils.attr_utils import (
    convert_sorted_kwargs_tuple,
)
from xorq.common.utils.func_utils import (
    return_constant,
)
from xorq.common.utils.otel_utils import (
    otel_config,
)


default_log_path = (
    Path(otel_config.OTEL_HOST_LOG_DIR, otel_config.OTEL_LOG_FILE_NAME)
    .expanduser()
    .resolve()
)

ANY = type(
    "ALL", (), {"__eq__": return_constant(True), "__contains__": return_constant(True)}
)()
NONE = type(
    "NONE",
    (),
    {"__eq__": return_constant(False), "__contains__": return_constant(False)},
)()


def dissoc_get(dct, *keys):
    values = tuple(dct.get(key) for key in keys)
    rest = toolz.dissoc(dct, *keys)
    return (values, rest)


datetime_from_unix_nano_str = compose(
    datetime.datetime.fromtimestamp,
    toolz.curried.flip(operator.truediv)(1e9),
    int,
)


def dissoc_get_all(dct, *keys):
    (values, rest) = dissoc_get(dct, *keys)
    assert not rest
    return values


def process_dict(dct, from_to_f):
    values = dissoc_get_all(dct, *from_to_f)
    processed = {
        to_: f(value) for ((_, (to_, f)), value) in zip(from_to_f.items(), values)
    }
    return processed


required_attribute = {
    "key": "service.name",
    "value": {
        "stringValue": "xorq-observability-service",
    },
}
required_scope = {"name": "xorq.tracer"}


def process_value(dct):
    ((value_type, value),) = dct.items()
    match value_type:
        case "intValue":
            value = int(value)
        case "stringValue":
            assert isinstance(value, str)
        case "arrayValue":
            if value:
                (lst,) = dissoc_get_all(value, "values")
                value = tuple(map(process_value, lst))
            else:
                value = ()
        case _:
            raise ValueError(f"Unhandled type {value_type}")
    return (value_type, value)


@frozen
class Link:
    trace_id = field(validator=instance_of(str))
    span_id = field(validator=instance_of(str))
    flags = field(validator=instance_of(int))

    @classmethod
    def from_dict(cls, dct):
        from_to_f = {
            "traceId": ("trace_id", identity),
            "spanId": ("span_id", identity),
            "flags": ("flags", identity),
        }
        kwargs = process_dict(dct, from_to_f)
        return cls(**kwargs)


@frozen
class Attribute:
    name = field(validator=instance_of(str))
    value_type = field(validator=instance_of(str))
    value = field()

    @classmethod
    def from_dict(cls, dct):
        (name, value) = dissoc_get_all(dct, "key", "value")
        value_type, value = process_value(value)
        return cls(name, value_type, value)


@frozen
class Event:
    time = field(validator=instance_of(datetime.datetime))
    name = field(validator=instance_of(str))
    attributes = field(
        validator=deep_iterable(instance_of(Attribute), instance_of(tuple))
    )

    @classmethod
    def from_dict(cls, dct):
        from_to_f = {
            "timeUnixNano": ("time", datetime_from_unix_nano_str),
            "name": ("name", identity),
            "attributes": ("attributes", compose(tuple, cmap(Attribute.from_dict))),
        }
        return cls(**process_dict(dct, from_to_f))


@frozen
class Span:
    trace_id = field(validator=instance_of(str))
    span_id = field(validator=instance_of(str))
    parent_span_id = field(validator=instance_of(str))
    name = field(validator=instance_of(str))
    flags = field(validator=instance_of(int))
    kind = field(validator=instance_of(int))
    links = field(validator=deep_iterable(instance_of(Link), instance_of(tuple)))
    start_datetime = field(validator=instance_of(datetime.datetime))
    end_datetime = field(validator=instance_of(datetime.datetime))
    events = field(
        validator=optional(deep_iterable(instance_of(Event), instance_of(tuple)))
    )
    attributes = field(
        validator=optional(deep_iterable(instance_of(Attribute), instance_of(tuple)))
    )
    status = field()

    @property
    def duration(self):
        return (self.end_datetime - self.start_datetime).total_seconds()

    @property
    def cache_event_dct(self):
        if self.name == "cache.set_default":
            (event,) = self.events
            (attribute,) = event.attributes
            return {
                "duration": self.duration,
                "name": event.name,
                "key": attribute.value,
            }
        else:
            return None

    @property
    def is_cache_hit(self):
        return self.name == "cache.set_default" and any(
            event.name == "cache.hit" for event in self.events
        )

    @property
    def is_cache_miss(self):
        return self.name == "cache.set_default" and any(
            event.name == "cache.miss" for event in self.events
        )

    def get_depth(self, trace):
        lineage = trace.get_lineage(self.span_id)
        return len(lineage) - 1

    @classmethod
    def from_dict(cls, dct):
        # FIXME: ensure we get all the fields
        kwargs = {
            to_: dct[from_]
            for to_, from_ in (
                ("trace_id", "traceId"),
                ("span_id", "spanId"),
                ("parent_span_id", "parentSpanId"),
                ("name", "name"),
                ("flags", "flags"),
                ("kind", "kind"),
            )
        } | {
            to_: f(dct.get(from_, default))
            for (to_, (from_, default, f)) in (
                (
                    "status",
                    ("status", None, convert_sorted_kwargs_tuple),
                ),
                (
                    "links",
                    (
                        "links",
                        (),
                        compose(
                            tuple,
                            cmap(Link.from_dict),
                        ),
                    ),
                ),
                (
                    "events",
                    (
                        "events",
                        (),
                        compose(
                            tuple,
                            cmap(Event.from_dict),
                        ),
                    ),
                ),
                (
                    "attributes",
                    (
                        "attributes",
                        (),
                        compose(
                            tuple,
                            cmap(Attribute.from_dict),
                        ),
                    ),
                ),
                (
                    "start_datetime",
                    ("startTimeUnixNano", None, datetime_from_unix_nano_str),
                ),
                (
                    "end_datetime",
                    ("endTimeUnixNano", None, datetime_from_unix_nano_str),
                ),
            )
        }
        return cls(**kwargs)

    @classmethod
    def spans_from_line(cls, line):
        assert line
        dct = json.loads(line)
        (resource_spans,) = dissoc_get_all(dct, "resourceSpans")
        resources, scope_spanss = zip(
            *(
                dissoc_get_all(resource_span, "resource", "scopeSpans")
                for resource_span in resource_spans
            )
        )
        assert all(
            dissoc_get_all(resource, "attributes") == ([required_attribute],)
            for resource in resources
        )
        scope_spans = tuple(scope_span for (scope_span,) in scope_spanss)
        (scopes, spanss) = zip(
            *(
                dissoc_get_all(scope_span, "scope", "spans")
                for scope_span in scope_spans
            )
        )
        assert all(scope == required_scope for scope in scopes)
        return tuple(Span.from_dict(dct) for spans in spanss for dct in spans)


@frozen
class Trace:
    spans = field(validator=deep_iterable(instance_of(Span), instance_of(tuple)))

    def __attrs_post_init__(self):
        end_datetimes = [span.end_datetime for span in self.spans]
        if not sorted(end_datetimes) == end_datetimes:
            raise ValueError("span end_datetimes must be in sorted order")

    def combine_with(self, other):
        return Trace(
            tuple(
                sorted(
                    self.spans + other.spans,
                    key=operator.attrgetter("end_datetime"),
                )
            )
        )

    @property
    def closed(self):
        trace_ids = {span.trace_id for span in self.spans}
        closed_trace_ids = {
            span.trace_id for span in self.spans if not span.parent_span_id
        }
        return trace_ids == closed_trace_ids and bool(
            toolz.excepts(Exception, operator.attrgetter("trace_metrics"))(self)
        )

    @property
    @functools.cache
    def trace_id(self):
        dct = toolz.groupby(
            compose(bool, operator.attrgetter("links")),
            self.spans,
        )
        with_links = dct.get(True, ())
        without_links = dct.get(False, ())
        if without_links:
            trace_id, *rest = {span.trace_id for span in without_links}
            assert not rest
            if with_links:
                # FIXME: handle links within links
                for span in with_links:
                    assert any(link.trace_id == trace_id for link in span.links)
            return trace_id
        elif with_links:
            # for now, require that there only be one linked traceId
            trace_id, *rest = {
                link.trace_id for span in with_links for link in span.links
            }
            assert not rest
            return trace_id
        else:
            raise ValueError("trace has no spans with or without links")

    @property
    @functools.cache
    def parent_span(self):
        (parent_span, *rest) = (
            span for span in self.spans if not span.parent_span_id and not span.links
        )
        assert not rest
        return parent_span

    @staticmethod
    def calc_duration(span):
        return (span.end_datetime - span.start_datetime).total_seconds()

    @property
    def cache_event_dcts(self):
        return tuple(filter(None, (span.cache_event_dct for span in self.spans)))

    def get_lineage(self, span_id):
        dct = {span.span_id: span for span in self.spans}
        lineage = ()
        while span_id:
            span = dct[span_id]
            span_id = span.parent_span_id
            lineage += (span,)
        return lineage

    def get_depth(self, depth):
        return self.get_depths().get(depth, ())

    def get_duration_delta(self, parent_span_id, lossless_leafs=True):
        parent_span = next(
            span for span in self.spans if span.span_id == parent_span_id
        )
        parent_depth = parent_span.get_depth(self)
        child_spans = tuple(
            span
            for span in self.get_depth(parent_depth + 1)
            if span.parent_span_id == parent_span_id
            # metrics_recorder spans can come up in non-parent spans because true parent went out of context
            and span.name != "otel_instrument_reader.metrics_recorder"
            # and span.name != "to_pyarrow_batches"
        )
        if child_spans or not lossless_leafs:
            return parent_span.duration - sum(
                child_span.duration for child_span in child_spans
            )
        else:
            return 0

    @functools.cache
    def get_depths(self):
        spans = tuple(span for span in self.spans if span != self.parent_span)
        depths = {
            0: (self.parent_span,),
        }
        depth = 1
        while spans:
            parent_span_ids = {span.span_id for span in depths[depth - 1]}
            dct = toolz.groupby(
                lambda span: span.parent_span_id in parent_span_ids,  # noqa: B023
                spans,
            )
            at_depth = dct.get(True, ())
            if not at_depth:
                if rest := dct.get(False):
                    depths[-1] = rest
                break
            else:
                depths[depth] = at_depth
            spans = dct.get(False, ())
            depth += 1
        return depths

    @property
    def duration(self):
        return self.calc_duration(self.parent_span)

    @property
    def start_datetime(self):
        return self.parent_span.start_datetime

    def get_spans_named(self, name):
        return tuple(span for span in self.spans if span.name == name)

    @property
    def trace_metrics(self):
        return TraceMetrics(self)

    @classmethod
    def accrue_spans(cls, spans, partials=()):
        def combine_traces(traces):
            def combine(traces):
                (trace, *others) = traces
                for other in others:
                    trace = trace.combine_with(other)
                return trace

            dct = toolz.valmap(
                combine,
                toolz.groupby(operator.attrgetter("trace_id"), traces),
            )
            return tuple(dct.values())

        def split_by_closed(traces):
            dct = toolz.groupby(operator.attrgetter("closed"), traces)
            closed = tuple(dct.get(True, ()))
            not_closed = tuple(dct.get(False, ()))
            return (closed, not_closed)

        maybe_partials = (
            tuple(
                toolz.valmap(
                    compose(Trace, tuple),
                    toolz.groupby(
                        operator.attrgetter("trace_id"),
                        spans,
                    ),
                ).values()
            )
            + partials
        )
        (traces, partials) = split_by_closed(combine_traces(maybe_partials))
        return (traces, partials)

    @classmethod
    def process_line(cls, line, partials=()):
        spans = Span.spans_from_line(line.strip())
        return cls.accrue_spans(spans, partials)

    @classmethod
    def process_lines(cls, lines, partials=()):
        spans = (span for line in lines for span in Span.spans_from_line(line))
        return cls.accrue_spans(spans, partials)

    @classmethod
    def process_path(cls, path=default_log_path):
        lines = path.read_text().strip().split("\n")
        (traces, partials) = cls.process_lines(lines)
        return (traces, partials)


@frozen
class TraceMetric:
    name = field(validator=instance_of(str))
    f = field(validator=is_callable())

    def calc_metric(self, trace):
        return self.f(trace)

    @classmethod
    def from_name_key_operator(cls, name, key, operator=sum):
        def f(trace):
            value = operator(
                attribute.value
                for el in trace.get_spans_named(name)
                for event in el.events
                for attribute in event.attributes
                if attribute.name == key
            )
            return value

        return cls(key, f)


def get_cache_metric(trace):
    # FIXME: deal with exception: has multiple attributes
    def process_span(span):
        assert span.name == "cache.set_default"
        (event,) = span.events
        (attribute,) = event.attributes
        return (event.name, attribute.value)

    return tuple(
        process_span(span) for span in trace.spans if span.name == "cache.set_default"
    )


n_batches_metric = TraceMetric.from_name_key_operator(
    "otel_instrument_reader.metrics_recorder", "n_batches"
)
sum_buffer_size_metric = TraceMetric.from_name_key_operator(
    "otel_instrument_reader.metrics_recorder", "sum_buffer_size"
)
cache_hit_metric = TraceMetric("trace_metrics", get_cache_metric)


@frozen
class TraceMetrics:
    trace = field(validator=instance_of(Trace))
    trace_metrics = field(
        validator=deep_iterable(instance_of(TraceMetric), instance_of(tuple)),
        default=(n_batches_metric, sum_buffer_size_metric, cache_hit_metric),
    )

    def __attrs_post_init__(self):
        self.check_validity()

    def check_validity(self):
        (oir, mr) = (
            self.trace.get_spans_named(name)
            for name in (
                "otel_instrument_reader",
                "otel_instrument_reader.metrics_recorder",
            )
        )
        assert len(oir) == len(mr)
        oir_ids, mr_ids = (
            {
                attribute.value
                for el in which
                for event in el.events
                for attribute in event.attributes
                if attribute.name == "reader_id"
            }
            for which in (oir, mr)
        )
        assert oir_ids == mr_ids

    @property
    @functools.cache
    def default_metrics(self):
        return {
            "trace_id": self.trace.trace_id,
            "duration": self.trace.duration,
            "start_datetime": self.trace.start_datetime,
        }

    @property
    @functools.cache
    def metrics(self):
        dct = {
            metric.name: metric.calc_metric(self.trace) for metric in self.trace_metrics
        } | self.default_metrics
        return dct


@frozen
class FileTailer:
    path = field(validator=instance_of(Path))
    sleep_duration = field(validator=instance_of(int), default=1)
    _fh = field(init=False)

    def __attrs_post_init__(self):
        if not self.path.exists():
            raise ValueError(f"path does not exist: {self.path}")
        fh = self.path.open("r")
        object.__setattr__(self, "_fh", fh)

    def get_line(self):
        content = self._fh.readline()
        while not content or not content[-1] == "\n":
            time.sleep(self.sleep_duration)
            content += self._fh.readline()
        return content

    def __iter__(self):
        return self

    def __next__(self):
        return self.get_line()


@frozen
class TraceFileProcessor:
    path = field(
        validator=instance_of(Path),
        default=default_log_path,
    )
    file_tailer = field(validator=instance_of(FileTailer), init=False)
    f = field(validator=is_callable(), default=toolz.identity)
    queue = field(validator=instance_of(Queue), init=False, factory=Queue)
    thread = field(validator=instance_of(Thread), init=False)

    def __attrs_post_init__(self):
        object.__setattr__(self, "file_tailer", FileTailer(self.path))
        object.__setattr__(self, "thread", Thread(target=self._run, daemon=True))
        self.thread.start()

    def _run(self):
        partials = ()
        for line in self.file_tailer:
            (traces, partials) = Trace.process_line(line, partials=partials)
            if traces is not None:
                for trace in traces:
                    self.queue.put(self.f(trace))

    def gen_existing(self):
        while True:
            try:
                yield self.queue.get(block=False)
            except QueueEmpty:
                break


def main():
    TraceFileProcessor(f=print)


if __name__ == "__main__":
    import sys

    sys.exit(main())



from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

from public import public

import xorq.common.exceptions as com
import xorq.vendor.ibis.expr.builders as bl
import xorq.vendor.ibis.expr.datatypes as dt
import xorq.vendor.ibis.expr.operations as ops
from xorq.common.utils.otel_utils import tracer
from xorq.vendor import ibis
from xorq.vendor.ibis.common.deferred import Deferred, _, deferrable
from xorq.vendor.ibis.common.grounds import Singleton
from xorq.vendor.ibis.expr.rewrites import rewrite_window_input
from xorq.vendor.ibis.expr.types.core import (
    Expr,
    _binop,
    _FixedTextJupyterMixin,
    _is_null_literal,
)
from xorq.vendor.ibis.util import deprecated, promote_list, warn_deprecated


if TYPE_CHECKING:
    import pandas as pd
    import polars as pl
    import pyarrow as pa
    import rich.table

    import xorq.vendor.ibis.expr.schema as sch
    import xorq.vendor.ibis.expr.types as ir
    from xorq.vendor.ibis.formats.pandas import PandasData
    from xorq.vendor.ibis.formats.pyarrow import PyArrowData


_SENTINEL = object()


@public
class Value(Expr):
    """Base class for a data generating expression having a known type."""

    def name(self, name):
        """Rename an expression to `name`.

        Parameters
        ----------
        name
            The new name of the expression

        Returns
        -------
        Value
            `self` with name `name`

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"a": [1, 2]})
        >>> t.a
        ┏━━━━━━━┓
        ┃ a     ┃
        ┡━━━━━━━┩
        │ int64 │
        ├───────┤
        │     1 │
        │     2 │
        └───────┘
        >>> t.a.name("b")
        ┏━━━━━━━┓
        ┃ b     ┃
        ┡━━━━━━━┩
        │ int64 │
        ├───────┤
        │     1 │
        │     2 │
        └───────┘
        """
        # TODO(kszucs): shouldn't do simplification here, but rather later
        # when simplifying the whole operation tree
        # the expression's name is idendical to the new one
        if self.get_name() == name:
            return self

        if isinstance(self.op(), ops.Alias):
            # only keep a single alias operation
            op = ops.Alias(arg=self.op().arg, name=name)
        else:
            op = ops.Alias(arg=self, name=name)

        return op.to_expr()

    # TODO(kszucs): should rename to dtype
    def type(self) -> dt.DataType:
        """Return the [DataType](./datatypes.qmd) of `self`."""
        return self.op().dtype

    def hash(self) -> ir.IntegerValue:
        """Compute an integer hash value.

        ::: {.callout-note}
        ## The hashing function used is backend-dependent.
        :::

        Returns
        -------
        IntegerValue
            The hash value of `self`

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> ibis.literal("hello").hash()  # doctest: +SKIP
        -4155090522938856779
        """
        return ops.Hash(self).to_expr()

    def cast(self, target_type: Any) -> Value:
        """Cast expression to indicated data type.

        Similar to `pandas.Series.astype`.

        Parameters
        ----------
        target_type
            Type to cast to. Anything accepted by [`ibis.dtype()`](./datatypes.qmd#ibis.dtype)

        Returns
        -------
        Value
            Casted expression

        See Also
        --------
        [`Value.try_cast()`](./expression-generic.qmd#ibis.expr.types.generic.Value.try_cast)
        [`ibis.dtype()`](./datatypes.qmd#ibis.dtype)

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> x = ibis.examples.penguins.fetch()["bill_depth_mm"]
        >>> x
        ┏━━━━━━━━━━━━━━━┓
        ┃ bill_depth_mm ┃
        ┡━━━━━━━━━━━━━━━┩
        │ float64       │
        ├───────────────┤
        │          18.7 │
        │          17.4 │
        │          18.0 │
        │          NULL │
        │          19.3 │
        │          20.6 │
        │          17.8 │
        │          19.6 │
        │          18.1 │
        │          20.2 │
        │             … │
        └───────────────┘

        python's built-in types can be used

        >>> x.cast(int)
        ┏━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ Cast(bill_depth_mm, int64) ┃
        ┡━━━━━━━━━━━━━━━━━━━━━━━━━━━━┩
        │ int64                      │
        ├────────────────────────────┤
        │                         19 │
        │                         17 │
        │                         18 │
        │                       NULL │
        │                         19 │
        │                         21 │
        │                         18 │
        │                         20 │
        │                         18 │
        │                         20 │
        │                          … │
        └────────────────────────────┘

        or string names

        >>> x.cast("uint16")
        ┏━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ Cast(bill_depth_mm, uint16) ┃
        ┡━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┩
        │ uint16                      │
        ├─────────────────────────────┤
        │                          19 │
        │                          17 │
        │                          18 │
        │                        NULL │
        │                          19 │
        │                          21 │
        │                          18 │
        │                          20 │
        │                          18 │
        │                          20 │
        │                           … │
        └─────────────────────────────┘

        If you make an illegal cast, you won't know until the backend actually
        executes it. Consider [`.try_cast()`](#ibis.expr.types.generic.Value.try_cast).

        >>> ibis.literal("a string").cast("int64")  # quartodoc: +SKIP # doctest: +SKIP
        <error>
        """
        op = ops.Cast(self, to=target_type)

        if op.to == self.type():
            # noop case if passed type is the same
            return self

        if op.to.is_geospatial() and not self.type().is_binary():
            from_geotype = self.type().geotype or "geometry"
            to_geotype = op.to.geotype
            if from_geotype == to_geotype:
                return self

        return op.to_expr()

    def try_cast(self, target_type: Any) -> Value:
        """Try cast expression to indicated data type.

        If the cast fails for a row, the value is returned
        as null or NaN depending on target_type and backend behavior.

        Parameters
        ----------
        target_type
            Type to try cast to. Anything accepted by [`ibis.dtype()`](./datatypes.qmd#ibis.dtype)

        Returns
        -------
        Value
            Casted expression

        See Also
        --------
        [`Value.cast()`](./expression-generic.qmd#ibis.expr.types.generic.Value.cast)
        [`ibis.dtype()`](./datatypes.qmd#ibis.dtype)

        Examples
        --------
        >>> import ibis
        >>> from ibis import _
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"numbers": [1, 2, 3, 4], "strings": ["1.0", "2", "hello", "world"]})
        >>> t
        ┏━━━━━━━━━┳━━━━━━━━━┓
        ┃ numbers ┃ strings ┃
        ┡━━━━━━━━━╇━━━━━━━━━┩
        │ int64   │ string  │
        ├─────────┼─────────┤
        │       1 │ 1.0     │
        │       2 │ 2       │
        │       3 │ hello   │
        │       4 │ world   │
        └─────────┴─────────┘
        >>> t = t.mutate(numbers_to_strings=_.numbers.try_cast("string"))
        >>> t = t.mutate(strings_to_numbers=_.strings.try_cast("int"))
        >>> t
        ┏━━━━━━━━━┳━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━┓
        ┃ numbers ┃ strings ┃ numbers_to_strings ┃ strings_to_numbers ┃
        ┡━━━━━━━━━╇━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━┩
        │ int64   │ string  │ string             │ int64              │
        ├─────────┼─────────┼────────────────────┼────────────────────┤
        │       1 │ 1.0     │ 1                  │                  1 │
        │       2 │ 2       │ 2                  │                  2 │
        │       3 │ hello   │ 3                  │               NULL │
        │       4 │ world   │ 4                  │               NULL │
        └─────────┴─────────┴────────────────────┴────────────────────┘
        """
        op = ops.TryCast(self, to=target_type)

        if op.to == self.type():
            # noop case if passed type is the same
            return self

        return op.to_expr()

    def coalesce(self, *args: Value) -> Value:
        """Return the first non-null value from `args`.

        Parameters
        ----------
        args
            Arguments from which to choose the first non-null value

        Returns
        -------
        Value
            Coalesced expression

        See Also
        --------
        [`ibis.coalesce()`](./expression-generic.qmd#ibis.coalesce)
        [`Value.fill_null()`](./expression-generic.qmd#ibis.expr.types.generic.Value.fill_null)

        Examples
        --------
        >>> import ibis
        >>> ibis.coalesce(None, 4, 5).name("x")
        x: Coalesce(...)
        """
        return ops.Coalesce((self, *args)).to_expr()

    @deprecated(as_of="8.0.0", instead="use ibis.greatest(self, rest...) instead")
    def greatest(self, *args: ir.Value) -> ir.Value:
        return ops.Greatest((self, *args)).to_expr()

    @deprecated(as_of="8.0.0", instead="use ibis.least(self, rest...) instead")
    def least(self, *args: ir.Value) -> ir.Value:
        return ops.Least((self, *args)).to_expr()

    def typeof(self) -> ir.StringValue:
        """Return the string name of the datatype of self.

        The values of the returned strings are necessarily backend dependent.
        e.g. duckdb may say "DOUBLE", while sqlite may say "real".

        Returns
        -------
        StringValue
            A string indicating the type of the value

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> vals = ibis.examples.penguins.fetch().head(5).bill_length_mm
        >>> vals
        ┏━━━━━━━━━━━━━━━━┓
        ┃ bill_length_mm ┃
        ┡━━━━━━━━━━━━━━━━┩
        │ float64        │
        ├────────────────┤
        │           39.1 │
        │           39.5 │
        │           40.3 │
        │           NULL │
        │           36.7 │
        └────────────────┘
        >>> vals.typeof()
        ┏━━━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ TypeOf(bill_length_mm) ┃
        ┡━━━━━━━━━━━━━━━━━━━━━━━━┩
        │ string                 │
        ├────────────────────────┤
        │ DOUBLE                 │
        │ DOUBLE                 │
        │ DOUBLE                 │
        │ DOUBLE                 │
        │ DOUBLE                 │
        └────────────────────────┘

        Different backends have different names for their native types

        >>> ibis.duckdb.connect().execute(ibis.literal(5.4).typeof())
        'DECIMAL(2,1)'
        >>> ibis.sqlite.connect().execute(ibis.literal(5.4).typeof())
        'real'
        """
        return ops.TypeOf(self).to_expr()

    def fill_null(self, fill_value: Scalar) -> Value:
        """Replace any null values with the indicated fill value.

        Parameters
        ----------
        fill_value
            Value with which to replace `NULL` values in `self`

        See Also
        --------
        [`Value.coalesce()`](./expression-generic.qmd#ibis.expr.types.generic.Value.coalesce)
        [`ibis.coalesce()`](./expression-generic.qmd#ibis.coalesce)

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.examples.penguins.fetch().limit(5)
        >>> t.sex
        ┏━━━━━━━━┓
        ┃ sex    ┃
        ┡━━━━━━━━┩
        │ string │
        ├────────┤
        │ male   │
        │ female │
        │ female │
        │ NULL   │
        │ female │
        └────────┘
        >>> t.sex.fill_null("unrecorded").name("sex")
        ┏━━━━━━━━━━━━┓
        ┃ sex        ┃
        ┡━━━━━━━━━━━━┩
        │ string     │
        ├────────────┤
        │ male       │
        │ female     │
        │ female     │
        │ unrecorded │
        │ female     │
        └────────────┘

        Returns
        -------
        Value
            `self` filled with `fill_value` where it is `NULL`
        """
        return ops.Coalesce((self, fill_value)).to_expr()

    @deprecated(as_of="9.1", instead="use fill_null instead")
    def fillna(self, fill_value: Scalar) -> Value:
        """Deprecated - use `fill_null` instead."""
        return self.fill_null(fill_value)

    def nullif(self, null_if_expr: Value) -> Value:
        """Set values to null if they equal the values `null_if_expr`.

        Commonly used to avoid divide-by-zero problems by replacing zero with
        `NULL` in the divisor.

        Equivalent to `(self == null_if_expr).ifelse(ibis.null(), self)`.

        Parameters
        ----------
        null_if_expr
            Expression indicating what values should be NULL

        Returns
        -------
        Value
            Value expression

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> vals = ibis.examples.penguins.fetch().head(5).sex
        >>> vals
        ┏━━━━━━━━┓
        ┃ sex    ┃
        ┡━━━━━━━━┩
        │ string │
        ├────────┤
        │ male   │
        │ female │
        │ female │
        │ NULL   │
        │ female │
        └────────┘
        >>> vals.nullif("male")
        ┏━━━━━━━━━━━━━━━━━━━━━┓
        ┃ NullIf(sex, 'male') ┃
        ┡━━━━━━━━━━━━━━━━━━━━━┩
        │ string              │
        ├─────────────────────┤
        │ NULL                │
        │ female              │
        │ female              │
        │ NULL                │
        │ female              │
        └─────────────────────┘
        """
        return ops.NullIf(self, null_if_expr).to_expr()

    def between(
        self,
        lower: Value,
        upper: Value,
    ) -> ir.BooleanValue:
        """Check if this expression is between `lower` and `upper`, inclusive.

        Parameters
        ----------
        lower
            Lower bound, inclusive
        upper
            Upper bound, inclusive

        Returns
        -------
        BooleanValue
            Expression indicating membership in the provided range

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.examples.penguins.fetch().limit(5)
        >>> t.bill_length_mm.between(35, 38)
        ┏━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ Between(bill_length_mm, 35, 38) ┃
        ┡━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┩
        │ boolean                         │
        ├─────────────────────────────────┤
        │ False                           │
        │ False                           │
        │ False                           │
        │ NULL                            │
        │ True                            │
        └─────────────────────────────────┘
        """
        return ops.Between(self, lower, upper).to_expr()

    def isin(self, values: Value | Sequence[Value]) -> ir.BooleanValue:
        """Check whether this expression's values are in `values`.

        `NULL` values are propagated in the output. See examples for details.

        Parameters
        ----------
        values
            Values or expression to check for membership

        Returns
        -------
        BooleanValue
            Expression indicating membership

        See Also
        --------
        [`Value.notin()`](./expression-generic.qmd#ibis.expr.types.generic.Value.notin)

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"a": [1, 2, 3], "b": [2, 3, 4]})
        >>> t
        ┏━━━━━━━┳━━━━━━━┓
        ┃ a     ┃ b     ┃
        ┡━━━━━━━╇━━━━━━━┩
        │ int64 │ int64 │
        ├───────┼───────┤
        │     1 │     2 │
        │     2 │     3 │
        │     3 │     4 │
        └───────┴───────┘

        Check against a literal sequence of values

        >>> t.a.isin([1, 2])
        ┏━━━━━━━━━━━━━━━━━━━━━┓
        ┃ InValues(a, (1, 2)) ┃
        ┡━━━━━━━━━━━━━━━━━━━━━┩
        │ boolean             │
        ├─────────────────────┤
        │ True                │
        │ True                │
        │ False               │
        └─────────────────────┘

        Check against a derived expression

        >>> t.a.isin(t.b + 1)
        ┏━━━━━━━━━━━━━━━┓
        ┃ InSubquery(a) ┃
        ┡━━━━━━━━━━━━━━━┩
        │ boolean       │
        ├───────────────┤
        │ False         │
        │ False         │
        │ True          │
        └───────────────┘

        Check against a column from a different table

        >>> t2 = ibis.memtable({"x": [99, 2, 99]})
        >>> t.a.isin(t2.x)
        ┏━━━━━━━━━━━━━━━┓
        ┃ InSubquery(a) ┃
        ┡━━━━━━━━━━━━━━━┩
        │ boolean       │
        ├───────────────┤
        │ False         │
        │ True          │
        │ False         │
        └───────────────┘

        `NULL` behavior

        >>> t = ibis.memtable({"x": [1, 2]})
        >>> t.x.isin([1, None])
        ┏━━━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ InValues(x, (1, None)) ┃
        ┡━━━━━━━━━━━━━━━━━━━━━━━━┩
        │ boolean                │
        ├────────────────────────┤
        │ True                   │
        │ NULL                   │
        └────────────────────────┘
        >>> t = ibis.memtable({"x": [1, None, 2]})
        >>> t.x.isin([1])
        ┏━━━━━━━━━━━━━━━━━━━┓
        ┃ InValues(x, (1,)) ┃
        ┡━━━━━━━━━━━━━━━━━━━┩
        │ boolean           │
        ├───────────────────┤
        │ True              │
        │ NULL              │
        │ False             │
        └───────────────────┘
        >>> t.x.isin([3])
        ┏━━━━━━━━━━━━━━━━━━━┓
        ┃ InValues(x, (3,)) ┃
        ┡━━━━━━━━━━━━━━━━━━━┩
        │ boolean           │
        ├───────────────────┤
        │ False             │
        │ NULL              │
        │ False             │
        └───────────────────┘
        """
        from xorq.vendor.ibis.expr.types import ArrayValue

        if isinstance(values, ArrayValue):
            return ops.ArrayContains(values, self).to_expr()
        elif isinstance(values, Column):
            return ops.InSubquery(values.as_table(), needle=self).to_expr()
        else:
            return ops.InValues(self, values).to_expr()

    def notin(self, values: Value | Sequence[Value]) -> ir.BooleanValue:
        """Check whether this expression's values are not in `values`.

        Opposite of [`Value.isin()`](./expression-generic.qmd#ibis.expr.types.generic.Value.isin).

        Parameters
        ----------
        values
            Values or expression to check for lack of membership

        Returns
        -------
        BooleanValue
            Whether `self`'s values are not contained in `values`

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.examples.penguins.fetch().limit(5)
        >>> t.bill_depth_mm
        ┏━━━━━━━━━━━━━━━┓
        ┃ bill_depth_mm ┃
        ┡━━━━━━━━━━━━━━━┩
        │ float64       │
        ├───────────────┤
        │          18.7 │
        │          17.4 │
        │          18.0 │
        │          NULL │
        │          19.3 │
        └───────────────┘
        >>> t.bill_depth_mm.notin([18.7, 18.1])
        ┏━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ Not(InValues(bill_depth_mm, (18.7, 18.1))) ┃
        ┡━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┩
        │ boolean                                    │
        ├────────────────────────────────────────────┤
        │ False                                      │
        │ True                                       │
        │ True                                       │
        │ NULL                                       │
        │ True                                       │
        └────────────────────────────────────────────┘
        """
        return ~self.isin(values)

    def substitute(
        self,
        value: Value | dict,
        replacement: Value | None = None,
        else_: Value | None = None,
    ):
        """Replace values given in `values` with `replacement`.

        This is similar to the pandas `replace` method.

        Parameters
        ----------
        value
            Expression or dict.
        replacement
            If an expression is passed to value, this must be
            passed.
        else_
            If an original value does not match `value`, then `else_` is used.
            The default of `None` means leave the original value unchanged.

        Returns
        -------
        Value
            Replaced values

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.examples.penguins.fetch()
        >>> t.island.value_counts().order_by("island")
        ┏━━━━━━━━━━━┳━━━━━━━━━━━━━━┓
        ┃ island    ┃ island_count ┃
        ┡━━━━━━━━━━━╇━━━━━━━━━━━━━━┩
        │ string    │ int64        │
        ├───────────┼──────────────┤
        │ Biscoe    │          168 │
        │ Dream     │          124 │
        │ Torgersen │           52 │
        └───────────┴──────────────┘
        >>> t.island.substitute({"Torgersen": "torg", "Biscoe": "bisc"}).name(
        ...     "island"
        ... ).value_counts().order_by("island")
        ┏━━━━━━━━┳━━━━━━━━━━━━━━┓
        ┃ island ┃ island_count ┃
        ┡━━━━━━━━╇━━━━━━━━━━━━━━┩
        │ string │ int64        │
        ├────────┼──────────────┤
        │ Dream  │          124 │
        │ bisc   │          168 │
        │ torg   │           52 │
        └────────┴──────────────┘
        """
        try:
            branches = value.items()
        except AttributeError:
            branches = [(value, replacement)]

        if (
            repl := next((v for k, v in branches if k is None), _SENTINEL)
        ) is not _SENTINEL:
            result = self.fill_null(repl)
        else:
            result = self

        if else_ is None:
            else_ = result

        if not (nonnulls := [(k, v) for k, v in branches if k is not None]):
            return else_

        return result.cases(*nonnulls, else_=else_)

    def over(
        self,
        window=None,
        *,
        rows=None,
        range=None,
        group_by=None,
        order_by=None,
    ) -> Value:
        """Construct a window expression.

        Parameters
        ----------
        window
            Window specification
        rows
            Whether to use the `ROWS` window clause
        range
            Whether to use the `RANGE` window clause
        group_by
            Grouping key
        order_by
            Ordering key

        Returns
        -------
        Value
            A window function expression

        """

        if window is None:
            window = ibis.window(
                rows=rows,
                range=range,
                group_by=group_by,
                order_by=order_by,
            )
        elif not isinstance(window, bl.WindowBuilder):
            raise com.XorqTypeError(f"Unexpected window type: {window!r}")

        node = self.op()
        if len(node.relations) == 0:
            table = None
        elif len(node.relations) == 1:
            (table,) = node.relations
            table = table.to_expr()
        else:
            raise com.RelationError("Cannot use window with multiple tables")

        @deferrable
        def bind(table):
            winfunc = rewrite_window_input(node, window.bind(table))
            if winfunc == node:
                raise com.XorqTypeError(
                    "No reduction or analytic function found to construct a window expression"
                )
            return winfunc.to_expr()

        try:
            return bind(table)
        except com.XorqInputError:
            return bind(_)

    def isnull(self) -> ir.BooleanValue:
        """Return whether this expression is NULL.

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.examples.penguins.fetch().limit(5)
        >>> t.bill_depth_mm
        ┏━━━━━━━━━━━━━━━┓
        ┃ bill_depth_mm ┃
        ┡━━━━━━━━━━━━━━━┩
        │ float64       │
        ├───────────────┤
        │          18.7 │
        │          17.4 │
        │          18.0 │
        │          NULL │
        │          19.3 │
        └───────────────┘
        >>> t.bill_depth_mm.isnull()
        ┏━━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ IsNull(bill_depth_mm) ┃
        ┡━━━━━━━━━━━━━━━━━━━━━━━┩
        │ boolean               │
        ├───────────────────────┤
        │ False                 │
        │ False                 │
        │ False                 │
        │ True                  │
        │ False                 │
        └───────────────────────┘
        """
        return ops.IsNull(self).to_expr()

    def notnull(self) -> ir.BooleanValue:
        """Return whether this expression is not NULL.

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.examples.penguins.fetch().limit(5)
        >>> t.bill_depth_mm
        ┏━━━━━━━━━━━━━━━┓
        ┃ bill_depth_mm ┃
        ┡━━━━━━━━━━━━━━━┩
        │ float64       │
        ├───────────────┤
        │          18.7 │
        │          17.4 │
        │          18.0 │
        │          NULL │
        │          19.3 │
        └───────────────┘
        >>> t.bill_depth_mm.notnull()
        ┏━━━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ NotNull(bill_depth_mm) ┃
        ┡━━━━━━━━━━━━━━━━━━━━━━━━┩
        │ boolean                │
        ├────────────────────────┤
        │ True                   │
        │ True                   │
        │ True                   │
        │ False                  │
        │ True                   │
        └────────────────────────┘
        """
        return ops.NotNull(self).to_expr()

    def case(self) -> bl.SimpleCaseBuilder:
        """Create a SimpleCaseBuilder to chain multiple if-else statements.

        Add new search expressions with the `.when()` method. These must be
        comparable with this column expression. Conclude by calling `.end()`.

        Returns
        -------
        SimpleCaseBuilder
            A case builder

        See Also
        --------
        [`Value.substitute()`](./expression-generic.qmd#ibis.expr.types.generic.Value.substitute)
        [`ibis.cases()`](./expression-generic.qmd#ibis.expr.types.generic.Value.cases)
        [`ibis.case()`](./expression-generic.qmd#ibis.case)

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> x = ibis.examples.penguins.fetch().head(5)["sex"]
        >>> x
        ┏━━━━━━━━┓
        ┃ sex    ┃
        ┡━━━━━━━━┩
        │ string │
        ├────────┤
        │ male   │
        │ female │
        │ female │
        │ NULL   │
        │ female │
        └────────┘
        >>> x.case().when("male", "M").when("female", "F").else_("U").end()
        ┏━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ SimpleCase(sex, ('male', 'female'), ('M', 'F'), 'U') ┃
        ┡━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┩
        │ string                                               │
        ├──────────────────────────────────────────────────────┤
        │ M                                                    │
        │ F                                                    │
        │ F                                                    │
        │ U                                                    │
        │ F                                                    │
        └──────────────────────────────────────────────────────┘

        Cases not given result in the ELSE case

        >>> x.case().when("male", "M").else_("OTHER").end()
        ┏━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ SimpleCase(sex, ('male',), ('M',), 'OTHER') ┃
        ┡━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┩
        │ string                                      │
        ├─────────────────────────────────────────────┤
        │ M                                           │
        │ OTHER                                       │
        │ OTHER                                       │
        │ OTHER                                       │
        │ OTHER                                       │
        └─────────────────────────────────────────────┘

        If you don't supply an ELSE, then NULL is used

        >>> x.case().when("male", "M").end()
        ┏━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ SimpleCase(sex, ('male',), ('M',), Cast(None, string)) ┃
        ┡━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┩
        │ string                                                 │
        ├────────────────────────────────────────────────────────┤
        │ M                                                      │
        │ NULL                                                   │
        │ NULL                                                   │
        │ NULL                                                   │
        │ NULL                                                   │
        └────────────────────────────────────────────────────────┘
        """
        import xorq.vendor.ibis.expr.builders as bl

        return bl.SimpleCaseBuilder(self.op())

    def cases(
        self,
        branch: tuple[Value, Value],
        /,
        *branches: tuple[Value, Value],
        else_: Value | None = None,
    ) -> Value:
        """Create a multi-branch if-else expression.

        Equivalent to a SQL `CASE` statement.

        Parameters
        ----------
        branch
            First (`condition`, `result`) pair. Required.
        branches
            Additional (`condition`, `result`) pairs. We look through the test
            values in order and return the result corresponding to the first
            test value that matches `self`. If none match, we return `else_`.
        else_
            Value to return if none of the case conditions evaluate to `True`.
            Defaults to `NULL`.

        Returns
        -------
        Value
            A value expression

        See Also
        --------
        [`Value.substitute()`](./expression-generic.qmd#ibis.expr.types.generic.Value.substitute)
        [`ibis.cases()`](./expression-generic.qmd#ibis.cases)

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable(
        ...     {
        ...         "left": [5, 6, 7, 8, 9, 10],
        ...         "symbol": ["+", "-", "*", "/", "bogus", None],
        ...         "right": [1, 2, 3, 4, 5, 6],
        ...     }
        ... )

        Note that we never hit the `None` case, because `x = NULL` is always
        `NULL`, which is not truthy. If you want to replace `NULL`s, you should use
        `.fill_null(some_value)` prior to `cases()`.

        >>> t.mutate(
        ...     result=(
        ...         t.symbol.cases(
        ...             ("+", t.left + t.right),
        ...             ("-", t.left - t.right),
        ...             ("*", t.left * t.right),
        ...             ("/", t.left / t.right),
        ...             (None, -999),
        ...         )
        ...     )
        ... )
        ┏━━━━━━━┳━━━━━━━━┳━━━━━━━┳━━━━━━━━━┓
        ┃ left  ┃ symbol ┃ right ┃ result  ┃
        ┡━━━━━━━╇━━━━━━━━╇━━━━━━━╇━━━━━━━━━┩
        │ int64 │ string │ int64 │ float64 │
        ├───────┼────────┼───────┼─────────┤
        │     5 │ +      │     1 │     6.0 │
        │     6 │ -      │     2 │     4.0 │
        │     7 │ *      │     3 │    21.0 │
        │     8 │ /      │     4 │     2.0 │
        │     9 │ bogus  │     5 │    NULL │
        │    10 │ NULL   │     6 │    NULL │
        └───────┴────────┴───────┴─────────┘
        """
        cases, results = zip(branch, *branches)
        return ops.SimpleCase(
            base=self, cases=cases, results=results, default=else_
        ).to_expr()

    def collect(
        self,
        where: ir.BooleanValue | None = None,
        order_by: Any = None,
        include_null: bool = False,
    ) -> ir.ArrayScalar:
        """Aggregate this expression's elements into an array.

        This function is called `array_agg`, `list_agg`, or `list` in other systems.

        Parameters
        ----------
        where
            An optional filter expression. If provided, only rows where `where`
            is `True` will be included in the aggregate.
        order_by
            An ordering key (or keys) to use to order the rows before
            aggregating. If not provided, the order of the items in the result
            is undefined and backend specific.
        include_null
            Whether to include null values when performing this aggregation. Set
            to `True` to include nulls in the result.

        Returns
        -------
        ArrayScalar
            Collected array

        Examples
        --------
        Basic collect usage

        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"key": list("aaabb"), "value": [1, 2, 3, 4, 5]})
        >>> t
        ┏━━━━━━━━┳━━━━━━━┓
        ┃ key    ┃ value ┃
        ┡━━━━━━━━╇━━━━━━━┩
        │ string │ int64 │
        ├────────┼───────┤
        │ a      │     1 │
        │ a      │     2 │
        │ a      │     3 │
        │ b      │     4 │
        │ b      │     5 │
        └────────┴───────┘
        >>> t.value.collect()
        ┌────────────────┐
        │ [1, 2, ... +3] │
        └────────────────┘
        >>> type(t.value.collect())
        <class 'ibis.expr.types.arrays.ArrayScalar'>

        Collect elements per group

        >>> t.group_by("key").agg(v=lambda t: t.value.collect()).order_by("key")
        ┏━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ key    ┃ v                    ┃
        ┡━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━┩
        │ string │ array<int64>         │
        ├────────┼──────────────────────┤
        │ a      │ [1, 2, ... +1]       │
        │ b      │ [4, 5]               │
        └────────┴──────────────────────┘

        Collect elements per group using a filter

        >>> t.group_by("key").agg(v=lambda t: t.value.collect(where=t.value > 1)).order_by("key")
        ┏━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ key    ┃ v                    ┃
        ┡━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━┩
        │ string │ array<int64>         │
        ├────────┼──────────────────────┤
        │ a      │ [2, 3]               │
        │ b      │ [4, 5]               │
        └────────┴──────────────────────┘
        """
        return ops.ArrayCollect(
            self,
            where=self._bind_to_parent_table(where),
            order_by=self._bind_order_by(order_by),
            include_null=include_null,
        ).to_expr()

    def identical_to(self, other: Value) -> ir.BooleanValue:
        """Return whether this expression is identical to other.

        Corresponds to `IS NOT DISTINCT FROM` in SQL.

        Parameters
        ----------
        other
            Expression to compare to

        Returns
        -------
        BooleanValue
            Whether this expression is not distinct from `other`

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> one = ibis.literal(1)
        >>> two = ibis.literal(2)
        >>> two.identical_to(one + one)
        ┌──────┐
        │ True │
        └──────┘
        """
        try:
            return ops.IdenticalTo(self, other).to_expr()
        except (com.XorqTypeError, NotImplementedError):
            return NotImplemented

    def group_concat(
        self,
        sep: str = ",",
        where: ir.BooleanValue | None = None,
        order_by: Any = None,
    ) -> ir.StringScalar:
        """Concatenate values using the indicated separator to produce a string.

        Parameters
        ----------
        sep
            The separator to use to join strings.
        where
            An optional filter expression. If provided, only rows where `where`
            is `True` will be included in the aggregate.
        order_by
            An ordering key (or keys) to use to order the rows before
            aggregating. If not provided, the order of the items in the result
            is undefined and backend specific.

        Returns
        -------
        StringScalar
            Concatenated string expression

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.examples.penguins.fetch().limit(5)
        >>> t[["bill_length_mm", "bill_depth_mm"]]
        ┏━━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━┓
        ┃ bill_length_mm ┃ bill_depth_mm ┃
        ┡━━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━┩
        │ float64        │ float64       │
        ├────────────────┼───────────────┤
        │           39.1 │          18.7 │
        │           39.5 │          17.4 │
        │           40.3 │          18.0 │
        │           NULL │          NULL │
        │           36.7 │          19.3 │
        └────────────────┴───────────────┘
        >>> t.bill_length_mm.group_concat()
        ┌─────────────────────┐
        │ 39.1,39.5,40.3,36.7 │
        └─────────────────────┘

        >>> t.bill_length_mm.group_concat(sep=": ")
        ┌────────────────────────┐
        │ 39.1: 39.5: 40.3: 36.7 │
        └────────────────────────┘

        >>> t.bill_length_mm.group_concat(sep=": ", where=t.bill_depth_mm > 18)
        ┌────────────┐
        │ 39.1: 36.7 │
        └────────────┘
        """
        return ops.GroupConcat(
            self,
            sep=sep,
            where=self._bind_to_parent_table(where),
            order_by=self._bind_order_by(order_by),
        ).to_expr()

    def __hash__(self) -> int:
        return super().__hash__()

    def __eq__(self, other: Value) -> ir.BooleanValue:
        if _is_null_literal(other):
            return self.isnull()
        elif _is_null_literal(self):
            return other.isnull()
        return _binop(ops.Equals, self, other)

    def __ne__(self, other: Value) -> ir.BooleanValue:
        if _is_null_literal(other):
            return self.notnull()
        elif _is_null_literal(self):
            return other.notnull()
        return _binop(ops.NotEquals, self, other)

    def __ge__(self, other: Value) -> ir.BooleanValue:
        return _binop(ops.GreaterEqual, self, other)

    def __gt__(self, other: Value) -> ir.BooleanValue:
        return _binop(ops.Greater, self, other)

    def __le__(self, other: Value) -> ir.BooleanValue:
        return _binop(ops.LessEqual, self, other)

    def __lt__(self, other: Value) -> ir.BooleanValue:
        return _binop(ops.Less, self, other)

    def asc(self, nulls_first: bool = False) -> ir.Value:
        """Sort an expression ascending."""
        return ops.SortKey(self, ascending=True, nulls_first=nulls_first).to_expr()

    def desc(self, nulls_first: bool = False) -> ir.Value:
        """Sort an expression descending."""
        return ops.SortKey(self, ascending=False, nulls_first=nulls_first).to_expr()

    def to_pandas(self, **kwargs) -> pd.Series:
        """Convert a column expression to a pandas Series or scalar object.

        Parameters
        ----------
        kwargs
            Same as keyword arguments to [`execute`](#ibis.expr.types.core.Expr.execute)

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.examples.penguins.fetch().limit(5)
        >>> t.to_pandas()
          species     island  bill_length_mm  ...  body_mass_g     sex  year
        0  Adelie  Torgersen            39.1  ...       3750.0    male  2007
        1  Adelie  Torgersen            39.5  ...       3800.0  female  2007
        2  Adelie  Torgersen            40.3  ...       3250.0  female  2007
        3  Adelie  Torgersen             NaN  ...          NaN    None  2007
        4  Adelie  Torgersen            36.7  ...       3450.0  female  2007
        [5 rows x 8 columns]
        """
        return self.execute(**kwargs)


@public
class Scalar(Value):
    def __pyarrow_result__(
        self,
        table: pa.Table,
        *,
        schema: sch.Schema | None = None,
        data_mapper: type[PyArrowData] | None = None,
    ) -> pa.Scalar:
        if data_mapper is None:
            from xorq.vendor.ibis.formats.pyarrow import PyArrowData as data_mapper

        return data_mapper.convert_scalar(
            table[0][0], self.type() if schema is None else schema.types[0]
        )

    @tracer.start_as_current_span("__pyarrow_result__")
    def __pandas_result__(
        self,
        df: pd.DataFrame,
        *,
        schema: sch.Schema | None = None,
        data_mapper: type[PandasData] | None = None,
    ) -> Any:
        if data_mapper is None:
            from xorq.vendor.ibis.formats.pandas import PandasData as data_mapper

        return data_mapper.convert_scalar(
            df, self.type() if schema is None else schema.types[0]
        )

    def __polars_result__(self, df: pl.DataFrame) -> Any:
        from xorq.vendor.ibis.formats.polars import PolarsData

        return PolarsData.convert_scalar(df, self.type())

    def as_scalar(self):
        """Inform ibis that the expression should be treated as a scalar.

        If the expression is a literal, it will be returned as is. If it depends
        on a table, it will be turned to a scalar subquery.

        Returns
        -------
        Scalar
            A scalar subquery or a literal

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.examples.penguins.fetch()
        >>> max_gentoo_weight = t.filter(t.species == "Gentoo").body_mass_g.max()
        >>> light_penguins = t.filter(t.body_mass_g < max_gentoo_weight / 2)
        >>> light_penguins.species.value_counts().order_by("species")
        ┏━━━━━━━━━━━┳━━━━━━━━━━━━━━━┓
        ┃ species   ┃ species_count ┃
        ┡━━━━━━━━━━━╇━━━━━━━━━━━━━━━┩
        │ string    │ int64         │
        ├───────────┼───────────────┤
        │ Adelie    │            15 │
        │ Chinstrap │             2 │
        └───────────┴───────────────┘
        """
        parents = self.op().relations
        if parents:
            return ops.ScalarSubquery(self.as_table()).to_expr()
        else:
            return self

    def as_table(self) -> ir.Table:
        """Promote the scalar expression to a table.

        Returns
        -------
        Table
            A table expression

        Examples
        --------
        Promote an aggregation to a table

        >>> import ibis
        >>> import ibis.expr.types as ir
        >>> t = ibis.table(dict(a="str"), name="t")
        >>> expr = t.a.length().sum().name("len").as_table()
        >>> isinstance(expr, ir.Table)
        True

        Promote a literal value to a table

        >>> import ibis.expr.types as ir
        >>> lit = ibis.literal(1).name("a").as_table()
        >>> isinstance(lit, ir.Table)
        True
        """
        from xorq.vendor.ibis.expr.types.relations import unwrap_alias

        op = self.op()
        parents = op.relations

        if not parents:
            return ops.DummyTable({op.name: unwrap_alias(op)}).to_expr()
        elif len(parents) == 1:
            (parent,) = parents
            return parent.to_expr().aggregate(self)
        else:
            raise com.RelationError(
                f"The scalar expression {self} cannot be converted to a "
                "table expression because it involves multiple base table "
                "references"
            )

    def __deferred_repr__(self):
        return f"<scalar[{self.type()}]>"

    def _repr_html_(self) -> str | None:
        return None


@public
class Column(Value, _FixedTextJupyterMixin):
    # Higher than numpy & dask objects
    __array_priority__ = 20

    __array_ufunc__ = None

    def __getitem__(self, _):
        raise TypeError(
            f"{self.__class__.__name__!r} is not subscriptable: "
            "see https://ibis-project.org/tutorial/ibis-for-pandas-users/#ibis-for-pandas-users for details."
        )

    def __array__(self, dtype=None):
        return self.execute().__array__(dtype)

    def preview(
        self,
        *,
        max_rows: int | None = None,
        max_length: int | None = None,
        max_string: int | None = None,
        max_depth: int | None = None,
        console_width: int | float | None = None,
    ) -> rich.table.Table:
        """Print a subset as a single-column Rich Table.

        This is an explicit version of what you get when you inspect
        this object in interactive mode, except with this version you
        can pass formatting options. The options are the same as those exposed
        in `ibis.options.interactive`.

        Parameters
        ----------
        max_rows
            Maximum number of rows to display
        max_length
            Maximum length for pretty-printed arrays and maps.
        max_string
            Maximum length for pretty-printed strings.
        max_depth
            Maximum depth for nested data types.
        console_width
            Width of the console in characters. If not specified, the width
            will be inferred from the console.

        Examples
        --------
        >>> import ibis
        >>> t = ibis.examples.penguins.fetch()
        >>> t.island.preview(max_rows=3, max_string=5)  # doctest: +SKIP
        ┏━━━━━━━━┓
        ┃ island ┃
        ┡━━━━━━━━┩
        │ stri…  │
        ├────────┤
        │ Torg…  │
        │ Torg…  │
        │ Torg…  │
        │ …      │
        └────────┘
        """
        from xorq.vendor.ibis.expr.types.pretty import to_rich

        return to_rich(
            self,
            max_rows=max_rows,
            max_length=max_length,
            max_string=max_string,
            max_depth=max_depth,
            console_width=console_width,
        )

    def __pyarrow_result__(
        self,
        table: pa.Table,
        *,
        schema: sch.Schema | None = None,
        data_mapper: type[PyArrowData] | None = None,
    ) -> pa.Array | pa.ChunkedArray:
        if data_mapper is None:
            from xorq.vendor.ibis.formats.pyarrow import PyArrowData as data_mapper

        return data_mapper.convert_column(
            table[0], self.type() if schema is None else schema.types[0]
        )

    def __pandas_result__(
        self,
        df: pd.DataFrame,
        *,
        schema: sch.Schema | None = None,
        data_mapper: type[PandasData] | None = None,
    ) -> pd.Series:
        if data_mapper is None:
            from xorq.vendor.ibis.formats.pandas import PandasData as data_mapper

        assert len(df.columns) == 1, (
            "more than one column when converting columnar result DataFrame to Series"
        )
        # in theory we could use df.iloc[:, 0], but there seems to be a bug in
        # older geopandas where df.iloc[:, 0] doesn't return the same kind of
        # object as df.loc[:, column_name] when df is a GeoDataFrame
        #
        # the bug is that iloc[:, 0] returns a bare series whereas
        # df.loc[:, column_name] returns the special GeoSeries object.
        #
        # this bug is fixed in later versions of geopandas
        return data_mapper.convert_column(
            df.loc[:, df.columns[0]], self.type() if schema is None else schema.types[0]
        )

    def __polars_result__(self, df: pl.DataFrame) -> pl.Series:
        from xorq.vendor.ibis.formats.polars import PolarsData

        return PolarsData.convert_column(df, self.type())

    def as_scalar(self) -> Scalar:
        """Inform ibis that the expression should be treated as a scalar.

        Creates a scalar subquery from the column expression. Since ibis cannot
        be sure that the column expression contains only one value, the column
        expression is wrapped in a scalar subquery and treated as a scalar.

        Note that the execution of the scalar subquery will fail if the column
        expression contains more than one value.

        Returns
        -------
        Scalar
            A scalar subquery

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.examples.penguins.fetch()
        >>> heavy_gentoo = t.filter(t.species == "Gentoo", t.body_mass_g > 6200)
        >>> from_that_island = t.filter(t.island == heavy_gentoo.island.as_scalar())
        >>> from_that_island.species.value_counts().order_by("species")
        ┏━━━━━━━━━┳━━━━━━━━━━━━━━━┓
        ┃ species ┃ species_count ┃
        ┡━━━━━━━━━╇━━━━━━━━━━━━━━━┩
        │ string  │ int64         │
        ├─────────┼───────────────┤
        │ Adelie  │            44 │
        │ Gentoo  │           124 │
        └─────────┴───────────────┘
        """
        return self.as_table().as_scalar()

    def as_table(self) -> ir.Table:
        """Promote the expression to a [Table](./expression-tables.qmd#ibis.expr.types.Table).

        Returns
        -------
        Table
            A table expression

        Examples
        --------
        >>> import ibis
        >>> t = ibis.table(dict(a="str"), name="t")
        >>> expr = t.a.length().name("len").as_table()
        >>> expected = t.select(len=t.a.length())
        >>> expr.equals(expected)
        True
        """
        from xorq.vendor.ibis.expr.types.relations import unwrap_alias

        op = self.op()
        parents = op.relations

        if not parents:
            return ops.DummyTable({op.name: unwrap_alias(op)}).to_expr()
        elif len(parents) == 1:
            (parent,) = parents
            return parent.to_expr().select(self)
        else:
            raise com.RelationError(
                f"Cannot convert {type(self)} expression involving multiple "
                "base table references to a projection"
            )

    def _bind_order_by(self, value) -> tuple[ops.SortKey, ...]:
        if value is None:
            return ()
        return tuple(self._bind_to_parent_table(v) for v in promote_list(value))

    def _bind_to_parent_table(self, value) -> Value | None:
        """Bind an expr to the parent table of `self`."""
        if value is None:
            return None
        if isinstance(value, (Deferred, str)) or callable(value):
            op = self.op()
            if len(op.relations) != 1:
                # TODO: I don't think this line can ever be hit by a valid
                # expression, since it would require a column expression to
                # directly depend on multiple tables. Currently some invalid
                # expressions (like t1.a.argmin(t2.b)) aren't caught at
                # construction time though, so we keep the check in for now.
                raise com.RelationError(
                    f"Unable to bind `{value!r}` - the current expression"
                    f"depends on multiple tables."
                )
            table = next(iter(op.relations)).to_expr()

            if isinstance(value, str):
                return table[value]
            elif isinstance(value, Deferred):
                return value.resolve(table)
            else:
                value = value(table)

        if not isinstance(value, Value):
            return literal(value)
        return value

    def __deferred_repr__(self):
        return f"<column[{self.type()}]>"

    def approx_nunique(self, where: ir.BooleanValue | None = None) -> ir.IntegerScalar:
        """Return the approximate number of distinct elements in `self`.

        ::: {.callout-note}
        ## The result may or may not be exact

        Whether the result is an approximation depends on the backend.

        ::: {.callout-warning}
        ## Do not depend on the results being exact
        :::

        :::

        Parameters
        ----------
        where
            Filter in values when `where` is `True`

        Returns
        -------
        Scalar
            An approximate count of the distinct elements of `self`

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.examples.penguins.fetch()
        >>> t.body_mass_g.approx_nunique()
        ┌────┐
        │ 92 │
        └────┘
        >>> t.body_mass_g.approx_nunique(where=t.species == "Adelie")
        ┌────┐
        │ 61 │
        └────┘
        """
        return ops.ApproxCountDistinct(
            self, where=self._bind_to_parent_table(where)
        ).to_expr()

    def approx_median(self, where: ir.BooleanValue | None = None) -> Scalar:
        """Return an approximate of the median of `self`.

        ::: {.callout-note}
        ## The result may or may not be exact

        Whether the result is an approximation depends on the backend.

        ::: {.callout-warning}
        ## Do not depend on the results being exact
        :::

        :::

        Parameters
        ----------
        where
            Filter in values when `where` is `True`

        Returns
        -------
        Scalar
            An approximation of the median of `self`

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.examples.penguins.fetch()
        >>> t.body_mass_g.approx_median()
        ┌────────┐
        │ 4030.0 │
        └────────┘
        >>> t.body_mass_g.approx_median(where=t.species == "Chinstrap")
        ┌────────┐
        │ 3700.0 │
        └────────┘
        """
        return ops.ApproxMedian(self, where=self._bind_to_parent_table(where)).to_expr()

    def mode(self, where: ir.BooleanValue | None = None) -> Scalar:
        """Return the mode of a column.

        Parameters
        ----------
        where
            Filter in values when `where` is `True`

        Returns
        -------
        Scalar
            The mode of `self`

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.examples.penguins.fetch()
        >>> t.body_mass_g.mode()
        ┌──────┐
        │ 3800 │
        └──────┘
        >>> t.body_mass_g.mode(where=(t.species == "Gentoo") & (t.sex == "male"))
        ┌──────┐
        │ 5550 │
        └──────┘
        """
        return ops.Mode(self, where=self._bind_to_parent_table(where)).to_expr()

    def max(self, where: ir.BooleanValue | None = None) -> Scalar:
        """Return the maximum of a column.

        Parameters
        ----------
        where
            Filter in values when `where` is `True`

        Returns
        -------
        Scalar
            The maximum value in `self`

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.examples.penguins.fetch()
        >>> t.body_mass_g.max()
        ┌──────┐
        │ 6300 │
        └──────┘
        >>> t.body_mass_g.max(where=t.species == "Chinstrap")
        ┌──────┐
        │ 4800 │
        └──────┘
        """
        return ops.Max(self, where=self._bind_to_parent_table(where)).to_expr()

    def min(self, where: ir.BooleanValue | None = None) -> Scalar:
        """Return the minimum of a column.

        Parameters
        ----------
        where
            Filter in values when `where` is `True`

        Returns
        -------
        Scalar
            The minimum value in `self`

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.examples.penguins.fetch()
        >>> t.body_mass_g.min()
        ┌──────┐
        │ 2700 │
        └──────┘
        >>> t.body_mass_g.min(where=t.species == "Adelie")
        ┌──────┐
        │ 2850 │
        └──────┘
        """
        return ops.Min(self, where=self._bind_to_parent_table(where)).to_expr()

    def argmax(self, key: ir.Value, where: ir.BooleanValue | None = None) -> Scalar:
        """Return the value of `self` that maximizes `key`.

        Parameters
        ----------
        key
            Key to use for `max` computation.
        where
            Keep values when `where` is `True`

        Returns
        -------
        Scalar
            The value of `self` that maximizes `key`

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.examples.penguins.fetch()
        >>> t.species.argmax(t.body_mass_g)
        ┌────────┐
        │ Gentoo │
        └────────┘
        >>> t.species.argmax(t.body_mass_g, where=t.island == "Dream")
        ┌───────────┐
        │ Chinstrap │
        └───────────┘
        """
        return ops.ArgMax(
            self,
            key=self._bind_to_parent_table(key),
            where=self._bind_to_parent_table(where),
        ).to_expr()

    def argmin(self, key: ir.Value, where: ir.BooleanValue | None = None) -> Scalar:
        """Return the value of `self` that minimizes `key`.

        Parameters
        ----------
        key
            Key to use for `min` computation.
        where
            Keep values when `where` is `True`

        Returns
        -------
        Scalar
            The value of `self` that minimizes `key`

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.examples.penguins.fetch()
        >>> t.species.argmin(t.body_mass_g)
        ┌───────────┐
        │ Chinstrap │
        └───────────┘

        >>> t.species.argmin(t.body_mass_g, where=t.island == "Biscoe")
        ┌────────┐
        │ Adelie │
        └────────┘
        """
        return ops.ArgMin(
            self,
            key=self._bind_to_parent_table(key),
            where=self._bind_to_parent_table(where),
        ).to_expr()

    def median(self, where: ir.BooleanValue | None = None) -> Scalar:
        """Return the median of the column.

        Parameters
        ----------
        where
            Optional boolean expression. If given, only the values where
            `where` evaluates to true will be considered for the median.

        Returns
        -------
        Scalar
            Median of the column

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.examples.penguins.fetch()

        Compute the median of `bill_depth_mm`

        >>> t.bill_depth_mm.median()
        ┌──────┐
        │ 17.3 │
        └──────┘
        >>> t.group_by(t.species).agg(median_bill_depth=t.bill_depth_mm.median()).order_by(
        ...     ibis.desc("median_bill_depth")
        ... )
        ┏━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━┓
        ┃ species   ┃ median_bill_depth ┃
        ┡━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━┩
        │ string    │ float64           │
        ├───────────┼───────────────────┤
        │ Chinstrap │             18.45 │
        │ Adelie    │             18.40 │
        │ Gentoo    │             15.00 │
        └───────────┴───────────────────┘

        In addition to numeric types, any orderable non-numeric types such as
        strings and dates work with `median`.

        >>> t.group_by(t.island).agg(median_species=t.species.median()).order_by(
        ...     ibis.desc("median_species")
        ... )
        ┏━━━━━━━━━━━┳━━━━━━━━━━━━━━━━┓
        ┃ island    ┃ median_species ┃
        ┡━━━━━━━━━━━╇━━━━━━━━━━━━━━━━┩
        │ string    │ string         │
        ├───────────┼────────────────┤
        │ Biscoe    │ Gentoo         │
        │ Dream     │ Chinstrap      │
        │ Torgersen │ Adelie         │
        └───────────┴────────────────┘
        """
        return ops.Median(self, where=self._bind_to_parent_table(where)).to_expr()

    def quantile(
        self,
        quantile: float | ir.NumericValue | Sequence[ir.NumericValue | float],
        where: ir.BooleanValue | None = None,
    ) -> Scalar:
        """Return value at the given quantile.

        The output of this method is a continuous quantile if the input is
        numeric, otherwise the output is a discrete quantile.

        Parameters
        ----------
        quantile
            `0 <= quantile <= 1`, or an array of such values
            indicating the quantile or quantiles to compute
        where
            Boolean filter for input values

        Returns
        -------
        Scalar
            Quantile of the input

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.examples.penguins.fetch()

        Compute the 99th percentile of `bill_depth`

        >>> t.bill_depth_mm.quantile(0.99)
        ┌──────┐
        │ 21.1 │
        └──────┘
        >>> t.group_by(t.species).agg(p99_bill_depth=t.bill_depth_mm.quantile(0.99)).order_by(
        ...     ibis.desc("p99_bill_depth")
        ... )
        ┏━━━━━━━━━━━┳━━━━━━━━━━━━━━━━┓
        ┃ species   ┃ p99_bill_depth ┃
        ┡━━━━━━━━━━━╇━━━━━━━━━━━━━━━━┩
        │ string    │ float64        │
        ├───────────┼────────────────┤
        │ Adelie    │         21.200 │
        │ Chinstrap │         20.733 │
        │ Gentoo    │         17.256 │
        └───────────┴────────────────┘

        In addition to numeric types, any orderable non-numeric types such as
        strings and dates work with `quantile`.

        Let's compute the 99th percentile of the `species` column

        >>> t.group_by(t.island).agg(p99_species=t.species.quantile(0.99)).order_by(
        ...     ibis.desc("p99_species")
        ... )
        ┏━━━━━━━━━━━┳━━━━━━━━━━━━━┓
        ┃ island    ┃ p99_species ┃
        ┡━━━━━━━━━━━╇━━━━━━━━━━━━━┩
        │ string    │ string      │
        ├───────────┼─────────────┤
        │ Biscoe    │ Gentoo      │
        │ Dream     │ Chinstrap   │
        │ Torgersen │ Adelie      │
        └───────────┴─────────────┘
        """
        if isinstance(quantile, Sequence):
            op = ops.MultiQuantile
        else:
            op = ops.Quantile
        return op(self, quantile, where=self._bind_to_parent_table(where)).to_expr()

    def nunique(self, where: ir.BooleanValue | None = None) -> ir.IntegerScalar:
        """Compute the number of distinct rows in an expression.

        Parameters
        ----------
        where
            Filter expression

        Returns
        -------
        IntegerScalar
            Number of distinct elements in an expression

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.examples.penguins.fetch()
        >>> t.body_mass_g.nunique()
        ┌────┐
        │ 94 │
        └────┘
        >>> t.body_mass_g.nunique(where=t.species == "Adelie")
        ┌────┐
        │ 55 │
        └────┘
        """
        return ops.CountDistinct(
            self, where=self._bind_to_parent_table(where)
        ).to_expr()

    def topk(
        self, k: int, by: ir.Value | None = None, *, name: str | None = None
    ) -> ir.Table:
        """Return a "top k" expression.

        Computes a Table containing the top `k` values by a certain metric
        (defaults to count).

        Parameters
        ----------
        k
            The number of rows to return.
        by
            The metric to compute "top" by. Defaults to `count`.
        name
            The name to use for the metric column. A suitable name will be
            automatically generated if not provided.

        Returns
        -------
        Table
            The top `k` values.

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.examples.diamonds.fetch()

        Compute the top 3 diamond colors by frequency:

        >>> t.color.topk(3)
        ┏━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━┓
        ┃ color  ┃ CountStar(diamonds) ┃
        ┡━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━┩
        │ string │ int64               │
        ├────────┼─────────────────────┤
        │ G      │               11292 │
        │ E      │                9797 │
        │ F      │                9542 │
        └────────┴─────────────────────┘

        Compute the top 3 diamond colors by mean price:

        >>> t.color.topk(3, by=t.price.mean())
        ┏━━━━━━━━┳━━━━━━━━━━━━━┓
        ┃ color  ┃ Mean(price) ┃
        ┡━━━━━━━━╇━━━━━━━━━━━━━┩
        │ string │ float64     │
        ├────────┼─────────────┤
        │ J      │ 5323.818020 │
        │ I      │ 5091.874954 │
        │ H      │ 4486.669196 │
        └────────┴─────────────┘

        Compute the top 2 diamond colors by max carat:

        >>> t.color.topk(2, by=t.carat.max(), name="max_carat")
        ┏━━━━━━━━┳━━━━━━━━━━━┓
        ┃ color  ┃ max_carat ┃
        ┡━━━━━━━━╇━━━━━━━━━━━┩
        │ string │ float64   │
        ├────────┼───────────┤
        │ J      │      5.01 │
        │ H      │      4.13 │
        └────────┴───────────┘
        """
        from xorq.vendor.ibis.expr.types.relations import bind

        try:
            (table,) = self.op().relations
        except ValueError:
            raise com.XorqTypeError("TopK must depend on exactly one table.")

        table = table.to_expr()

        if by is None:
            by = lambda t: t.count()  # noqa: E731

        (metric,) = bind(table, by)

        if name is not None:
            metric = metric.name(name)

        return table.aggregate(metric, by=[self]).order_by(metric.desc()).limit(k)

    def arbitrary(
        self, where: ir.BooleanValue | None = None, how: Any = None
    ) -> Scalar:
        """Select an arbitrary value in a column.

        Returns an arbitrary (nondeterministic, backend-specific) value from
        the column. The value will be non-NULL, except if the column is empty
        or all values are NULL.

        Parameters
        ----------
        where
            A filter expression
        how
            DEPRECATED

        Returns
        -------
        Scalar
            An expression

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"a": [1, 2, 2], "b": list("aaa"), "c": [4.0, 4.1, 4.2]})
        >>> t
        ┏━━━━━━━┳━━━━━━━━┳━━━━━━━━━┓
        ┃ a     ┃ b      ┃ c       ┃
        ┡━━━━━━━╇━━━━━━━━╇━━━━━━━━━┩
        │ int64 │ string │ float64 │
        ├───────┼────────┼─────────┤
        │     1 │ a      │     4.0 │
        │     2 │ a      │     4.1 │
        │     2 │ a      │     4.2 │
        └───────┴────────┴─────────┘
        >>> t.group_by("a").agg(arb=t.b.arbitrary(), c=t.c.sum()).order_by("a")
        ┏━━━━━━━┳━━━━━━━━┳━━━━━━━━━┓
        ┃ a     ┃ arb    ┃ c       ┃
        ┡━━━━━━━╇━━━━━━━━╇━━━━━━━━━┩
        │ int64 │ string │ float64 │
        ├───────┼────────┼─────────┤
        │     1 │ a      │     4.0 │
        │     2 │ a      │     8.3 │
        └───────┴────────┴─────────┘
        """
        if how is not None:
            warn_deprecated(
                name="how",
                as_of="9.0",
                removed_in="10.0",
                instead="call `first` or `last` explicitly",
            )
        return ops.Arbitrary(self, where=self._bind_to_parent_table(where)).to_expr()

    def count(self, where: ir.BooleanValue | None = None) -> ir.IntegerScalar:
        """Compute the number of rows in an expression.

        Parameters
        ----------
        where
            Filter expression

        Returns
        -------
        IntegerScalar
            Number of elements in an expression
        """
        return ops.Count(self, where=self._bind_to_parent_table(where)).to_expr()

    def value_counts(self, *, name: str | None = None) -> ir.Table:
        """Compute a frequency table.

        Parameters
        ----------
        name
            The name to use for the frequency column. A suitable name will be
            automatically generated if not provided.

        Returns
        -------
        Table
            The frequency table.

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"chars": ["a", "a", "b", "c", "c", "c", "d", "d", "d", "d"]})

        Compute the count of each unique value in "chars", ordered by "chars":

        >>> t.chars.value_counts().order_by("chars")
        ┏━━━━━━━━┳━━━━━━━━━━━━━┓
        ┃ chars  ┃ chars_count ┃
        ┡━━━━━━━━╇━━━━━━━━━━━━━┩
        │ string │ int64       │
        ├────────┼─────────────┤
        │ a      │           2 │
        │ b      │           1 │
        │ c      │           3 │
        │ d      │           4 │
        └────────┴─────────────┘

        Compute the count of each unique value in "chars" as a column named
        "freq", ordered by "freq":

        >>> t.chars.value_counts(name="freq").order_by("freq")
        ┏━━━━━━━━┳━━━━━━━┓
        ┃ chars  ┃ freq  ┃
        ┡━━━━━━━━╇━━━━━━━┩
        │ string │ int64 │
        ├────────┼───────┤
        │ b      │     1 │
        │ a      │     2 │
        │ c      │     3 │
        │ d      │     4 │
        └────────┴───────┘
        """
        colname = self.get_name()
        if name is None:
            name = f"{colname}_count"
        t = self.as_table()
        return t.group_by(t[colname]).aggregate(t.count().name(name))

    def first(
        self,
        where: ir.BooleanValue | None = None,
        order_by: Any = None,
        include_null: bool = False,
    ) -> Value:
        """Return the first value of a column.

        Parameters
        ----------
        where
            An optional filter expression. If provided, only rows where `where`
            is `True` will be included in the aggregate.
        order_by
            An ordering key (or keys) to use to order the rows before
            aggregating. If not provided, the meaning of `first` is undefined
            and will be backend specific.
        include_null
            Whether to include null values when performing this aggregation. Set
            to `True` to include nulls in the result.

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"chars": ["a", "b", "c", "d"]})
        >>> t
        ┏━━━━━━━━┓
        ┃ chars  ┃
        ┡━━━━━━━━┩
        │ string │
        ├────────┤
        │ a      │
        │ b      │
        │ c      │
        │ d      │
        └────────┘
        >>> t.chars.first()
        ┌───┐
        │ a │
        └───┘
        >>> t.chars.first(where=t.chars != "a")
        ┌───┐
        │ b │
        └───┘
        """
        return ops.First(
            self,
            where=self._bind_to_parent_table(where),
            order_by=self._bind_order_by(order_by),
            include_null=include_null,
        ).to_expr()

    def last(
        self,
        where: ir.BooleanValue | None = None,
        order_by: Any = None,
        include_null: bool = False,
    ) -> Value:
        """Return the last value of a column.

        Parameters
        ----------
        where
            An optional filter expression. If provided, only rows where `where`
            is `True` will be included in the aggregate.
        order_by
            An ordering key (or keys) to use to order the rows before
            aggregating. If not provided, the meaning of `last` is undefined
            and will be backend specific.
        include_null
            Whether to include null values when performing this aggregation. Set
            to `True` to include nulls in the result.

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"chars": ["a", "b", "c", "d"]})
        >>> t
        ┏━━━━━━━━┓
        ┃ chars  ┃
        ┡━━━━━━━━┩
        │ string │
        ├────────┤
        │ a      │
        │ b      │
        │ c      │
        │ d      │
        └────────┘
        >>> t.chars.last()
        ┌───┐
        │ d │
        └───┘
        >>> t.chars.last(where=t.chars != "d")
        ┌───┐
        │ c │
        └───┘
        """
        return ops.Last(
            self,
            where=self._bind_to_parent_table(where),
            order_by=self._bind_order_by(order_by),
            include_null=include_null,
        ).to_expr()

    def rank(self) -> ir.IntegerColumn:
        """Compute position of first element within each equal-value group in sorted order.

        Equivalent to SQL's `RANK()` window function.

        Returns
        -------
        Int64Column
            The min rank

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"values": [1, 2, 1, 2, 3, 2]})
        >>> t.mutate(rank=t.values.rank())
        ┏━━━━━━━━┳━━━━━━━┓
        ┃ values ┃ rank  ┃
        ┡━━━━━━━━╇━━━━━━━┩
        │ int64  │ int64 │
        ├────────┼───────┤
        │      1 │     0 │
        │      1 │     0 │
        │      2 │     2 │
        │      2 │     2 │
        │      2 │     2 │
        │      3 │     5 │
        └────────┴───────┘
        """
        return ibis.rank().over(order_by=self)

    def dense_rank(self) -> ir.IntegerColumn:
        """Position of first element within each group of equal values.

        Values are returned in sorted order and duplicate values are ignored.

        Equivalent to SQL's `DENSE_RANK()`.

        Returns
        -------
        IntegerColumn
            The rank

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"values": [1, 2, 1, 2, 3, 2]})
        >>> t.mutate(rank=t.values.dense_rank())
        ┏━━━━━━━━┳━━━━━━━┓
        ┃ values ┃ rank  ┃
        ┡━━━━━━━━╇━━━━━━━┩
        │ int64  │ int64 │
        ├────────┼───────┤
        │      1 │     0 │
        │      1 │     0 │
        │      2 │     1 │
        │      2 │     1 │
        │      2 │     1 │
        │      3 │     2 │
        └────────┴───────┘
        """
        return ibis.dense_rank().over(order_by=self)

    def percent_rank(self) -> Column:
        """Return the relative rank of the values in the column."""
        return ibis.percent_rank().over(order_by=self)

    def cume_dist(self) -> Column:
        """Return the cumulative distribution over a window."""
        return ibis.cume_dist().over(order_by=self)

    def ntile(self, buckets: int | ir.IntegerValue) -> ir.IntegerColumn:
        """Return the integer number of a partitioning of the column values.

        Parameters
        ----------
        buckets
            Number of buckets to partition into
        """
        return ibis.ntile(buckets).over(order_by=self)

    def cummin(self, *, where=None, group_by=None, order_by=None) -> Column:
        """Return the cumulative min over a window."""
        return self.min(where=where).over(
            ibis.cumulative_window(group_by=group_by, order_by=order_by)
        )

    def cummax(self, *, where=None, group_by=None, order_by=None) -> Column:
        """Return the cumulative max over a window."""
        return self.max(where=where).over(
            ibis.cumulative_window(group_by=group_by, order_by=order_by)
        )

    def lag(
        self,
        offset: int | ir.IntegerValue | None = None,
        default: Value | None = None,
    ) -> Column:
        """Return the row located at `offset` rows **before** the current row.

        Parameters
        ----------
        offset
            Index of row to select
        default
            Value used if no row exists at `offset`
        """
        return ops.Lag(self, offset, default).to_expr()

    def lead(
        self,
        offset: int | ir.IntegerValue | None = None,
        default: Value | None = None,
    ) -> Column:
        """Return the row located at `offset` rows **after** the current row.

        Parameters
        ----------
        offset
            Index of row to select
        default
            Value used if no row exists at `offset`
        """
        return ops.Lead(self, offset, default).to_expr()

    def nth(self, n: int | ir.IntegerValue) -> Column:
        """Return the `n`th value (0-indexed) over a window.

        `.nth(0)` is equivalent to `.first()`. Negative will result in `NULL`.
        If the value of `n` is greater than the number of rows in the window,
        `NULL` will be returned.

        Parameters
        ----------
        n
            Desired rank value

        Returns
        -------
        Column
            The nth value over a window
        """
        return ops.NthValue(self, n).to_expr()


@public
class UnknownValue(Value):
    pass


@public
class UnknownScalar(Scalar):
    pass


@public
class UnknownColumn(Column):
    pass


@public
class NullValue(Value):
    pass


@public
class NullScalar(Scalar, NullValue, Singleton):
    pass


@public
class NullColumn(Column, NullValue):
    pass


@public
@deferrable
def null(type: dt.DataType | str | None = None) -> Value:
    """Create a NULL scalar.

    `NULL`s with an unspecified type are castable and comparable to values,
    but lack datatype-specific methods:

    >>> import ibis
    >>> ibis.options.interactive = True
    >>> ibis.null().upper()
    Traceback (most recent call last):
        ...
    AttributeError: 'NullScalar' object has no attribute 'upper'
    >>> ibis.null(str).upper()
    ┌──────┐
    │ None │
    └──────┘
    >>> ibis.null(str).upper().isnull()
    ┌──────┐
    │ True │
    └──────┘
    """
    if type is None:
        type = dt.null
    return ops.Literal(None, type).to_expr()


@public
@deferrable
def literal(value: Any, type: dt.DataType | str | None = None) -> Scalar:
    """Create a scalar expression from a Python value.

    ::: {.callout-tip}
    ## Use specific functions for arrays, structs and maps

    Ibis supports literal construction of arrays using the following
    functions:

    1. [`ibis.array`](./expression-collections.qmd#ibis.array)
    1. [`ibis.struct`](./expression-collections.qmd#ibis.struct)
    1. [`ibis.map`](./expression-collections.qmd#ibis.map)

    Constructing these types using `literal` will be deprecated in a future
    release.
    :::

    Parameters
    ----------
    value
        A Python value
    type
        An instance of [`DataType`](./datatypes.qmd#ibis.expr.datatypes.DataType) or a string
        indicating the ibis type of `value`. This parameter can be used
        in cases where ibis's type inference isn't sufficient for discovering
        the type of `value`.

    Returns
    -------
    Scalar
        An expression representing a literal value

    Examples
    --------
    Construct an integer literal

    >>> import ibis
    >>> x = ibis.literal(42)
    >>> x.type()
    Int8(nullable=True)

    Construct a `float64` literal from an `int`

    >>> y = ibis.literal(42, type="double")
    >>> y.type()
    Float64(nullable=True)

    Ibis checks for invalid types

    >>> ibis.literal("foobar", type="int64")  # quartodoc: +EXPECTED_FAILURE
    Traceback (most recent call last):
      ...
    TypeError: Value 'foobar' cannot be safely coerced to int64

    Literals can also be used in a deferred context.

    Here's an example of constructing a table of a column's type repeated for
    every row:

    >>> from ibis import _, selectors as s
    >>> ibis.options.interactive = True
    >>> t = ibis.examples.penguins.fetch()
    >>> t.select(s.across(s.all(), ibis.literal(_.type(), type=str).name(_.get_name()))).head(1)
    ┏━━━━━━━━━┳━━━━━━━━┳━━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━┳━━━┓
    ┃ species ┃ island ┃ bill_length_mm ┃ bill_depth_mm ┃ flipper_length_mm ┃ … ┃
    ┡━━━━━━━━━╇━━━━━━━━╇━━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━╇━━━┩
    │ string  │ string │ string         │ string        │ string            │ … │
    ├─────────┼────────┼────────────────┼───────────────┼───────────────────┼───┤
    │ string  │ string │ float64        │ float64       │ int64             │ … │
    └─────────┴────────┴────────────────┴───────────────┴───────────────────┴───┘
    """
    if isinstance(value, Expr):
        node = value.op()
        if not isinstance(node, ops.Literal):
            raise TypeError(f"Ibis expression {value!r} is not a Literal")
        if type is None or node.dtype.castable(dt.dtype(type)):
            return value
        else:
            raise TypeError(
                f"Ibis literal {value!r} cannot be safely coerced to datatype {type}"
            )

    dtype = dt.infer(value) if type is None else dt.dtype(type)
    return ops.Literal(value, dtype=dtype).to_expr()


public(
    ValueExpr=Value,
    ScalarExpr=Scalar,
    ColumnExpr=Column,
    AnyValue=Value,
    AnyScalar=Scalar,
    AnyColumn=Column,
)



import hashlib
import itertools
import pathlib
import pdb
from contextlib import contextmanager
from unittest.mock import (
    patch,
)

import dask
import toolz


def normalize_attrs(attrs):
    assert hasattr(attrs, "__attrs_attrs__")
    return tuple(sorted(attrs.__getstate__().items()))


@contextmanager
def patch_normalize_token(*typs, f):
    lazy_before = tuple(dask.base.normalize_token._lazy.items())
    try:
        with patch.dict(
            dask.base.normalize_token._lookup,
            values=dict.fromkeys(typs, f),
        ) as dct:
            yield dct
    finally:
        lazy_after = tuple(dask.base.normalize_token._lazy.items())
        registers = tuple(
            register for _, register in set(lazy_before).difference(lazy_after)
        )
        for register in registers:
            register()


def normalize_seq_with_caller(*args, caller=""):
    from xorq.common.utils.inspect_utils import (
        get_enclosing_function,
    )

    if caller is None:
        caller = get_enclosing_function(level=2)
    return dask.tokenize._normalize_seq_func(
        (
            caller,
            args,
        )
    )


@toolz.curry
def walk_normalized(f, normalized):
    match normalized:
        case tuple() | list():
            for el in normalized:
                yield from walk_normalized(f, el)
        case str():
            yield f(normalized)
        case bytes():
            yield f(normalized.decode("ascii", errors="replace"))
        case int() | float():
            yield f(normalized)
        case None | slice():
            yield f(normalized)
        case _:
            raise ValueError(
                f"unhandled condition for type {type(normalized)} ({normalized})"
            )


@toolz.curry
def set_trace_on_condition(condition, obj):
    if condition(obj):
        pdb.set_trace()


def gen_batches(path, size=2**20):
    with pathlib.Path(path).open("rb") as fh:
        gen = (fh.read(size) for fh in itertools.repeat(fh))
        gen = itertools.takewhile(bool, gen)
        yield from gen


def manual_file_digest(path, digest=hashlib.md5, size=2**20):
    from contextlib import closing
    from tarfile import ExFileObject

    fh = path if isinstance(path, ExFileObject) else pathlib.Path(path).open("rb")
    with closing(fh):
        obj = digest()
        for chunk in itertools.takewhile(
            bool, (fh.read(size) for fh in itertools.repeat(fh))
        ):
            obj.update(chunk)
        return obj.hexdigest()


def file_digest(path, digest=hashlib.md5, size=2**20):
    from tarfile import ExFileObject

    if hasattr(hashlib, "file_digest"):
        if isinstance(path, ExFileObject):
            return hashlib.file_digest(path, digest).hexdigest()
        elif isinstance(path, (str, pathlib.Path)):
            with pathlib.Path(path).open("rb") as fh:
                return hashlib.file_digest(fh, digest).hexdigest()
        else:
            raise ValueError(f"Don't know how to handle type {type}")
    else:
        # python 3.10
        return manual_file_digest(path, digest, size=size)


def normalize_read_path_md5sum(path):
    tpls = (("content-md5sum", file_digest(path)),)
    return tpls


@contextmanager
def patch_normalize_op_caching():
    import functools

    import xorq.common.utils.dask_normalize.dask_normalize_expr as mod

    attr = "normalize_op"
    cached_attr = functools.cache(getattr(mod, attr))
    with patch.object(mod, attr, cached_attr):
        yield


def patched_tokenize(expr):
    with patch_normalize_op_caching():
        return dask.base.tokenize(expr)



import time

import pandas as pd
import pyarrow as pa

from xorq.flight.metrics import instrument_rpc, setup_console_metrics


@instrument_rpc("dummy_do_get")
def dummy_do_get(self, context, *args, **kwargs):
    # Create a dictionary with column names as keys and lists as values
    data = {
        "Name": ["John", "Alice", "Bob"],
        "Age": [25, 30, 35],
        "City": ["New York", "London", "Paris"],
    }

    # Create the DataFrame
    df = pd.DataFrame(data)

    return pa.Table.from_pandas(df)


def test_setup_console_metrics_simple(capsys):
    setup_console_metrics()
    dummy_do_get(None, None)

    for _i in range(20):
        time.sleep(1)
        if (
            "flight_server.requests_total{method=dummy_do_get} 1"
            in capsys.readouterr().out
        ):
            return

    raise AssertionError()



import sys

import pytest
from click.testing import CliRunner

from xorq.cli import OutputFormats, cli


@pytest.mark.skipif(
    sys.version_info > (3, 10), reason="compatibility test for Python 3.10"
)
@pytest.mark.xorq
def test_output_formats_enum():
    for fmt in OutputFormats:
        assert OutputFormats(fmt.value) == fmt
    assert OutputFormats.default == OutputFormats.parquet


@pytest.mark.skipif(
    sys.version_info > (3, 10), reason="compatibility test for Python 3.10"
)
@pytest.mark.xorq
@pytest.mark.parametrize("cmd", ["run", "run-cached", "uv-run", "run-unbound"])
def test_output_format_choices_in_help(cmd):
    runner = CliRunner()
    result = runner.invoke(cli, [cmd, "--help"])
    assert result.exit_code == 0, result.output
    for fmt in OutputFormats:
        assert fmt.value in result.output


@pytest.mark.skipif(
    sys.version_info > (3, 10), reason="compatibility test for Python 3.10"
)
@pytest.mark.xorq
@pytest.mark.parametrize("fmt", OutputFormats)
def test_output_format_accepted_by_cli(tmp_path, fmt):
    runner = CliRunner()
    result = runner.invoke(cli, ["run", str(tmp_path), "--format", fmt.value])
    assert "Invalid value for '-f' / '--format'" not in (result.output or "")



from __future__ import annotations

import calendar
import math
from functools import partial
from itertools import starmap

import sqlglot as sg
import sqlglot.expressions as sge

import xorq.common.exceptions as com
import xorq.vendor.ibis.expr.datatypes as dt
import xorq.vendor.ibis.expr.operations as ops
from xorq.vendor.ibis.backends.sql.compilers.base import (
    FALSE,
    NULL,
    STAR,
    AggGen,
    SQLGlotCompiler,
)
from xorq.vendor.ibis.backends.sql.datatypes import DataFusionType
from xorq.vendor.ibis.backends.sql.dialects import DataFusion
from xorq.vendor.ibis.backends.sql.rewrites import split_select_distinct_with_order_by
from xorq.vendor.ibis.common.temporal import IntervalUnit, TimestampUnit
from xorq.vendor.ibis.expr.operations.udf import InputType


class DataFusionCompiler(SQLGlotCompiler):
    __slots__ = ()

    dialect = DataFusion
    type_mapper = DataFusionType

    agg = AggGen(supports_filter=True, supports_order_by=True)

    post_rewrites = (split_select_distinct_with_order_by,)

    UNSUPPORTED_OPS = (
        ops.ArgMax,
        ops.ArgMin,
        ops.ArrayDistinct,
        ops.ArrayFilter,
        ops.ArrayMap,
        ops.ArrayZip,
        ops.BitwiseNot,
        ops.Clip,
        ops.CountDistinctStar,
        ops.DateDelta,
        ops.Greatest,
        ops.IntervalFromInteger,
        ops.Least,
        ops.RowID,
        ops.Strftime,
        ops.TimeDelta,
        ops.TimestampBucket,
        ops.TimestampDelta,
        ops.TypeOf,
        ops.StringToDate,
        ops.StringToTimestamp,
    )

    SIMPLE_OPS = {
        ops.ApproxQuantile: "approx_percentile_cont",
        ops.ApproxMedian: "approx_median",
        ops.ArrayRemove: "array_remove_all",
        ops.BitAnd: "bit_and",
        ops.BitOr: "bit_or",
        ops.BitXor: "bit_xor",
        ops.Cot: "cot",
        ops.ExtractMicrosecond: "extract_microsecond",
        ops.Median: "median",
        ops.StringLength: "character_length",
        ops.RegexSplit: "regex_split",
        ops.EndsWith: "ends_with",
        ops.ArrayIntersect: "array_intersect",
        ops.ArrayUnion: "array_union",
    }

    def _to_timestamp(self, value, target_dtype, literal=False):
        tz = (
            f'Some("{timezone}")'
            if (timezone := target_dtype.timezone) is not None
            else "None"
        )
        unit = (
            target_dtype.unit.name.capitalize()
            if target_dtype.scale is not None
            else "Microsecond"
        )
        str_value = str(value) if literal else value
        return self.f.arrow_cast(str_value, f"Timestamp({unit}, {tz})")

    def visit_NonNullLiteral(self, op, *, value, dtype):
        if dtype.is_decimal():
            return self.cast(
                sg.exp.convert(str(value)),
                dt.Decimal(precision=dtype.precision or 38, scale=dtype.scale or 9),
            )
        elif dtype.is_numeric():
            if isinstance(value, float):
                if math.isinf(value):
                    return self.cast("+Inf", dt.float64)
                elif math.isnan(value):
                    return self.cast("NaN", dt.float64)
            return sg.exp.convert(value)
        elif dtype.is_interval():
            if dtype.unit.short in ("ms", "us", "ns"):
                raise com.UnsupportedOperationError(
                    "DataFusion doesn't support subsecond interval resolutions"
                )

            return sg.exp.Interval(
                this=sg.exp.convert(str(value)),
                unit=sg.exp.var(dtype.unit.plural.lower()),
            )
        elif dtype.is_timestamp():
            return self._to_timestamp(value, dtype, literal=True)
        elif dtype.is_date():
            return self.f.date_trunc("day", value.isoformat())
        elif dtype.is_binary():
            return sg.exp.HexString(this=value.hex())
        elif dtype.is_uuid():
            return sge.convert(str(value))
        elif dtype.is_struct():
            args = []
            for name, field_value in value.items():
                args.append(sge.convert(name))
                args.append(field_value)
            return self.f.named_struct(*args)
        else:
            return None

    def visit_Cast(self, op, *, arg, to):
        if to.is_interval():
            unit = to.unit.name.lower()
            return sg.cast(
                self.f.concat(self.cast(arg, dt.string), f" {unit}"), "interval"
            )
        if to.is_timestamp():
            return self._to_timestamp(arg, to)
        if to.is_decimal():
            from xorq.vendor.ibis.formats.pyarrow import PyArrowType

            return self.f.arrow_cast(arg, f"{PyArrowType.from_ibis(to)}".capitalize())
        return self.cast(arg, to)

    def visit_Arbitrary(self, op, *, arg, where):
        cond = ~arg.is_(NULL)
        if where is not None:
            cond &= where
        return self.agg.first_value(arg, where=cond)

    def visit_Variance(self, op, *, arg, how, where):
        if how == "sample":
            return self.agg.var_samp(arg, where=where)
        elif how == "pop":
            return self.agg.var_pop(arg, where=where)
        else:
            raise ValueError(f"Unrecognized how value: {how}")

    def visit_StandardDev(self, op, *, arg, how, where):
        if how == "sample":
            return self.agg.stddev_samp(arg, where=where)
        elif how == "pop":
            return self.agg.stddev_pop(arg, where=where)
        else:
            raise ValueError(f"Unrecognized how value: {how}")

    def visit_ScalarUDF(self, op, **kw):
        input_type = op.__input_type__
        if input_type in (InputType.PYARROW, InputType.BUILTIN):
            return self.f.anon[op.__func_name__](*kw.values())
        else:
            raise NotImplementedError(
                f"DataFusion only supports PyArrow UDFs: got a {input_type.name.lower()} UDF"
            )

    def visit_ElementWiseVectorizedUDF(
        self, op, *, func, func_args, input_type, return_type
    ):
        return self.f[func.__name__](*func_args)

    def visit_RegexExtract(self, op, *, arg, pattern, index):
        if not isinstance(op.index, ops.Literal):
            raise ValueError(
                "re_extract `index` expressions must be literals. "
                "Arbitrary expressions are not supported in the DataFusion backend"
            )
        return self.f.regexp_match(arg, self.f.concat("(", pattern, ")"))[index]

    def visit_StringFind(self, op, *, arg, substr, start, end):
        if end is not None:
            raise NotImplementedError("`end` not yet implemented")

        if start is not None:
            pos = self.f.strpos(self.f.substr(arg, start + 1), substr)
            return self.f.coalesce(self.f.nullif(pos + start, start), 0)

        return self.f.strpos(arg, substr)

    def visit_RegexSearch(self, op, *, arg, pattern):
        return self.if_(
            sg.or_(arg.is_(NULL), pattern.is_(NULL)),
            NULL,
            self.f.coalesce(
                # null is returned for non-matching patterns, so coalesce to false
                # because that is the desired behavior for ops.RegexSearch
                self.f.array_length(self.f.regexp_match(arg, pattern)) > 0,
                FALSE,
            ),
        )

    def visit_StringContains(self, op, *, haystack, needle):
        return self.f.strpos(haystack, needle) > sg.exp.convert(0)

    def visit_ExtractFragment(self, op, *, arg):
        return self.f.extract_url_field(arg, "fragment")

    def visit_ExtractProtocol(self, op, *, arg):
        return self.f.extract_url_field(arg, "scheme")

    def visit_ExtractAuthority(self, op, *, arg):
        return self.f.extract_url_field(arg, "netloc")

    def visit_ExtractPath(self, op, *, arg):
        return self.f.extract_url_field(arg, "path")

    def visit_ExtractHost(self, op, *, arg):
        return self.f.extract_url_field(arg, "hostname")

    def visit_ExtractQuery(self, op, *, arg, key):
        if key is not None:
            return self.f.extract_query_param(arg, key)
        return self.f.extract_query(arg)

    def visit_ExtractUserInfo(self, op, *, arg):
        return self.f.extract_user_info(arg)

    def visit_ExtractYearMonthQuarterDay(self, op, *, arg):
        skip = len("Extract")
        part = type(op).__name__[skip:].lower()
        return self.f.date_part(part, arg)

    visit_ExtractYear = visit_ExtractMonth = visit_ExtractQuarter = visit_ExtractDay = (
        visit_ExtractYearMonthQuarterDay
    )

    def visit_ExtractDayOfYear(self, op, *, arg):
        return self.f.date_part("doy", arg)

    def visit_DayOfWeekIndex(self, op, *, arg):
        return (self.f.date_part("dow", arg) + 6) % 7

    def visit_DayOfWeekName(self, op, *, arg):
        return sg.exp.Case(
            this=sge.paren(self.f.date_part("dow", arg) + 6, copy=False) % 7,
            ifs=list(starmap(self.if_, enumerate(calendar.day_name))),
        )

    def visit_Date(self, op, *, arg):
        return self.f.date_trunc("day", arg)

    def visit_ExtractWeekOfYear(self, op, *, arg):
        return self.f.date_part("week", arg)

    def visit_TimestampTruncate(self, op, *, arg, unit):
        if unit in (
            IntervalUnit.MILLISECOND,
            IntervalUnit.MICROSECOND,
            IntervalUnit.NANOSECOND,
        ):
            raise com.UnsupportedOperationError(
                f"The function is not defined for time unit {unit}"
            )

        return self.f.date_trunc(unit.name.lower(), arg)

    def visit_ExtractEpochSeconds(self, op, *, arg):
        if op.arg.dtype.is_date():
            return self.f.extract_epoch_seconds_date(arg)
        elif op.arg.dtype.is_timestamp():
            return self.f.extract_epoch_seconds_timestamp(arg)
        else:
            raise com.OperationNotDefinedError(
                f"The function is not defined for {op.arg.dtype}"
            )

    def visit_ExtractMinute(self, op, *, arg):
        if op.arg.dtype.is_date():
            return self.f.date_part("minute", arg)
        elif op.arg.dtype.is_time():
            return self.f.extract_minute_time(arg)
        elif op.arg.dtype.is_timestamp():
            return self.f.extract_minute_timestamp(arg)
        else:
            raise com.OperationNotDefinedError(
                f"The function is not defined for {op.arg.dtype}"
            )

    def visit_ExtractMillisecond(self, op, *, arg):
        if op.arg.dtype.is_time():
            return self.f.extract_millisecond_time(arg)
        elif op.arg.dtype.is_timestamp():
            return self.f.extract_millisecond_timestamp(arg)
        else:
            raise com.OperationNotDefinedError(
                f"The function is not defined for {op.arg.dtype}"
            )

    def visit_ExtractHour(self, op, *, arg):
        if op.arg.dtype.is_date() or op.arg.dtype.is_timestamp():
            return self.f.date_part("hour", arg)
        elif op.arg.dtype.is_time():
            return self.f.extract_hour_time(arg)
        else:
            raise com.OperationNotDefinedError(
                f"The function is not defined for {op.arg.dtype}"
            )

    def visit_ExtractSecond(self, op, *, arg):
        if op.arg.dtype.is_date() or op.arg.dtype.is_timestamp():
            return self.f.extract_second_timestamp(arg)
        elif op.arg.dtype.is_time():
            return self.f.extract_second_time(arg)
        else:
            raise com.OperationNotDefinedError(
                f"The function is not defined for {op.arg.dtype}"
            )

    def visit_ArrayRepeat(self, op, *, arg, times):
        return self.f.flatten(self.f.array_repeat(arg, times))

    def visit_ArrayPosition(self, op, *, arg, other):
        return self.f.coalesce(self.f.array_position(arg, other), 0)

    def visit_ArrayCollect(self, op, *, arg, where, order_by, include_null):
        if not include_null:
            cond = arg.is_(sg.not_(NULL, copy=False))
            where = cond if where is None else sge.And(this=cond, expression=where)
        return self.agg.array_agg(arg, where=where, order_by=order_by)

    def visit_Covariance(self, op, *, left, right, how, where):
        x = op.left
        if x.dtype.is_boolean():
            left = self.cast(left, dt.float64)

        y = op.right
        if y.dtype.is_boolean():
            right = self.cast(right, dt.float64)

        if how == "sample":
            return self.agg.covar_samp(left, right, where=where)
        elif how == "pop":
            return self.agg.covar_pop(left, right, where=where)
        else:
            raise ValueError(f"Unrecognized how = `{how}` value")

    def visit_Correlation(self, op, *, left, right, where, how):
        x = op.left
        if x.dtype.is_boolean():
            left = self.cast(left, dt.float64)

        y = op.right
        if y.dtype.is_boolean():
            right = self.cast(right, dt.float64)

        return self.agg.corr(left, right, where=where)

    def visit_IsNan(self, op, *, arg):
        return sg.and_(arg.is_(sg.not_(NULL)), self.f.isnan(arg))

    def visit_ArrayStringJoin(self, op, *, sep, arg):
        return self.f.array_join(arg, sep)

    def visit_FindInSet(self, op, *, needle, values):
        return self.f.coalesce(
            self.f.array_position(self.f.make_array(*values), needle), 0
        )

    def visit_TimestampFromUNIX(self, op, *, arg, unit):
        if unit == TimestampUnit.SECOND:
            return self.f.from_unixtime(arg)
        elif unit in (
            TimestampUnit.MILLISECOND,
            TimestampUnit.MICROSECOND,
            TimestampUnit.NANOSECOND,
        ):
            return self.f.arrow_cast(arg, f"Timestamp({unit.name.capitalize()}, None)")
        else:
            raise com.UnsupportedOperationError(f"Unsupported unit {unit}")

    def visit_DateFromYMD(self, op, *, year, month, day):
        return self.cast(
            self.f.concat(
                self.f.lpad(self.cast(self.cast(year, dt.int64), dt.string), 4, "0"),
                "-",
                self.f.lpad(self.cast(self.cast(month, dt.int64), dt.string), 2, "0"),
                "-",
                self.f.lpad(self.cast(self.cast(day, dt.int64), dt.string), 2, "0"),
            ),
            dt.date,
        )

    def visit_TimestampFromYMDHMS(
        self, op, *, year, month, day, hours, minutes, seconds, **_
    ):
        return self.f.to_timestamp_micros(
            self.f.concat(
                self.f.lpad(self.cast(self.cast(year, dt.int64), dt.string), 4, "0"),
                "-",
                self.f.lpad(self.cast(self.cast(month, dt.int64), dt.string), 2, "0"),
                "-",
                self.f.lpad(self.cast(self.cast(day, dt.int64), dt.string), 2, "0"),
                "T",
                self.f.lpad(self.cast(self.cast(hours, dt.int64), dt.string), 2, "0"),
                ":",
                self.f.lpad(self.cast(self.cast(minutes, dt.int64), dt.string), 2, "0"),
                ":",
                self.f.lpad(self.cast(self.cast(seconds, dt.int64), dt.string), 2, "0"),
                ".000000Z",
            )
        )

    def visit_IsInf(self, op, *, arg):
        return sg.and_(sg.not_(self.f.isnan(arg)), self.f.abs(arg).eq(self.POS_INF))

    def visit_ArrayIndex(self, op, *, arg, index):
        return self.f.array_element(arg, index + self.cast(index >= 0, op.index.dtype))

    def visit_StringConcat(self, op, *, arg):
        any_args_null = (a.is_(NULL) for a in arg)
        return self.if_(
            sg.or_(*any_args_null), self.cast(NULL, dt.string), self.f.concat(*arg)
        )

    def visit_First(self, op, *, arg, where, order_by, include_null):
        if not include_null:
            cond = arg.is_(sg.not_(NULL, copy=False))
            where = cond if where is None else sge.And(this=cond, expression=where)
        return self.agg.first_value(arg, where=where, order_by=order_by)

    def visit_Last(self, op, *, arg, where, order_by, include_null):
        if not include_null:
            cond = arg.is_(sg.not_(NULL, copy=False))
            where = cond if where is None else sge.And(this=cond, expression=where)
        return self.agg.last_value(arg, where=where, order_by=order_by)

    def visit_Aggregate(self, op, *, parent, groups, metrics):
        """Support `GROUP BY` expressions in `SELECT` since DataFusion does not."""
        quoted = self.quoted
        metrics = tuple(self._cleanup_names(metrics))

        if groups:
            # datafusion doesn't support count distinct aggregations alongside
            # computed grouping keys so create a projection of the key and all
            # existing columns first, followed by the usual group by
            #
            # analogous to a user calling mutate -> group_by
            cols = list(
                map(
                    partial(
                        sg.column,
                        table=sg.to_identifier(parent.alias, quoted=quoted),
                        quoted=quoted,
                    ),
                    # can't use set subtraction here since the schema keys'
                    # order matters and set subtraction doesn't preserve order
                    (k for k in op.parent.schema.keys() if k not in groups),
                )
            )
            table = (
                sg.select(*cols, *self._cleanup_names(groups))
                .from_(parent)
                .subquery(parent.alias)
            )

            # datafusion lower cases all column names internally unless quoted so
            # quoted=True is required here for correctness
            by_names_quoted = tuple(
                sg.column(key, table=getattr(value, "table", None), quoted=quoted)
                for key, value in groups.items()
            )
            selections = by_names_quoted + metrics
        else:
            selections = metrics or (STAR,)
            table = parent

        sel = sg.select(*selections).from_(table)

        if groups:
            sel = sel.group_by(*by_names_quoted)

        return sel

    def visit_StructColumn(self, op, *, names, values):
        args = []
        for name, value in zip(names, values):
            args.append(sge.convert(name))
            args.append(value)
        return self.f.named_struct(*args)

    def visit_GroupConcat(self, op, *, arg, sep, where, order_by):
        if order_by:
            raise com.UnsupportedOperationError(
                "DataFusion does not support order-sensitive group_concat"
            )
        return super().visit_GroupConcat(
            op, arg=arg, sep=sep, where=where, order_by=order_by
        )

    def visit_ArrayFlatten(self, op, *, arg):
        return self.if_(arg.is_(NULL), NULL, self.f.flatten(arg))


compiler = DataFusionCompiler()



from __future__ import annotations

import functools
from typing import TYPE_CHECKING, Any

from public import public

import xorq.vendor.ibis.expr.datatypes.core as dt
from xorq.common.exceptions import XorqTypeError


if TYPE_CHECKING:
    from collections.abc import Iterator


@public
def cast(source: str | dt.DataType, target: str | dt.DataType, **kwargs) -> dt.DataType:
    """Attempts to implicitly cast from source dtype to target dtype."""
    source, target = dt.dtype(source), dt.dtype(target)

    if not source.castable(target, **kwargs):
        raise XorqTypeError(
            f"Datatype {source} cannot be implicitly casted to {target}"
        )
    return target


@public
def castable(source: dt.DataType, target: dt.DataType, value: Any = None) -> bool:
    """Return whether source ir type is implicitly castable to target."""
    from xorq.vendor.ibis.expr.datatypes.value import normalizable

    if source == target:
        return True
    elif source.is_null():
        # The null type is castable to any type, even if the target type is *not*
        # nullable.
        #
        # We handle the promotion of `null + !T -> T` at the `castable` call site.
        #
        # It might be possible to build a system with a single function that tries
        # to promote types and use the exception to indicate castability, but that
        # is a deeper refactor to be tackled later.
        #
        # See https://github.com/ibis-project/ibis/issues/2891 for the bug report
        return True
    elif target.is_boolean():
        if source.is_boolean():
            return True
        elif source.is_integer():
            return value in (0, 1)
        else:
            return False
    elif target.is_integer():
        # TODO(kszucs): ideally unsigned to signed shouldn't be allowed but that
        # breaks the integral promotion rule logic in rules.py
        if source.is_integer():
            if value is not None:
                return normalizable(target, value)
            else:
                return source.nbytes <= target.nbytes
        else:
            return False
    elif target.is_floating():
        if source.is_floating():
            return source.nbytes <= target.nbytes
        else:
            return source.is_integer()
    elif target.is_decimal():
        if source.is_decimal():
            downcast_precision = (
                source.precision is not None
                and target.precision is not None
                and source.precision < target.precision
            )
            downcast_scale = (
                source.scale is not None
                and target.scale is not None
                and source.scale < target.scale
            )
            return not (downcast_precision or downcast_scale)
        else:
            return source.is_numeric()
    elif target.is_string():
        return source.is_string() or source.is_uuid()
    elif target.is_uuid():
        return source.is_uuid() or source.is_string()
    elif target.is_date() or target.is_timestamp():
        if source.is_string():
            return value is not None and normalizable(target, value)
        else:
            return source.is_timestamp() or source.is_date()
    elif target.is_interval():
        if source.is_interval():
            return source.unit == target.unit
        else:
            return source.is_integer()
    elif target.is_time():
        if source.is_string():
            return value is not None and normalizable(target, value)
        else:
            return source.is_time()
    elif target.is_json():
        return (
            source.is_json()
            or source.is_string()
            or source.is_floating()
            or source.is_integer()
        )
    elif target.is_array():
        return source.is_array() and castable(source.value_type, target.value_type)
    elif target.is_map():
        return (
            source.is_map()
            and castable(source.key_type, target.key_type)
            and castable(source.value_type, target.value_type)
        )
    elif target.is_struct():
        return source.is_struct() and all(
            castable(source[field], target[field]) for field in target.names
        )
    elif target.is_geospatial():
        return source.is_geospatial() or source.is_array()
    else:
        return isinstance(target, source.__class__)


@public
def higher_precedence(left: dt.DataType, right: dt.DataType) -> dt.DataType:
    nullable = left.nullable or right.nullable

    if left.castable(right):
        return right.copy(nullable=nullable)
    elif right.castable(left):
        return left.copy(nullable=nullable)
    else:
        raise XorqTypeError(
            f"Cannot compute precedence for `{left}` and `{right}` types"
        )


@public
def highest_precedence(dtypes: Iterator[dt.DataType]) -> dt.DataType:
    """Compute the highest precedence of `dtypes`."""
    if collected := list(dtypes):
        return functools.reduce(higher_precedence, collected)
    else:
        return dt.null
